# -*- coding: utf-8 -*-


import dask.array as da

from . import _utils


def _binary_op(func,
               image,
               structure=None,
               iterations=1,
               mask=None,
               origin=0,
               brute_force=False,
               **kwargs):
    image = (image != 0)

    structure = _utils._get_structure(image, structure)
    iterations = _utils._get_iterations(iterations)
    mask = _utils._get_mask(image, mask)
    origin = _utils._get_origin(structure.shape, origin)
    brute_force = _utils._get_brute_force(brute_force)
    depth = _utils._get_depth(structure.shape, origin)
    depth, boundary = _utils._get_depth_boundary(structure.ndim, depth, "none")

    result = image
    for i in range(iterations):
        iter_result = result.map_overlap(
            func,
            depth=depth,
            boundary=boundary,
            dtype=bool,
            meta=image._meta,
            structure=structure,
            origin=origin,
            **kwargs
        )
        result = da.where(mask, iter_result, result)
        result._meta = image._meta.astype(bool)

    return result



# -*- coding: utf-8 -*-


import scipy.ndimage

from ..dispatch._dispatch_ndfilters import dispatch_laplace
from . import _utils

__all__ = [
    "laplace",
]


@_utils._update_wrapper(scipy.ndimage.laplace)
def laplace(image, mode='reflect', cval=0.0):
    result = image.map_overlap(
        dispatch_laplace(image),
        depth=(image.ndim * (1,)),
        boundary="none",
        dtype=image.dtype,
        meta=image._meta,
        mode=mode,
        cval=cval
    )

    return result



# -*- coding: utf-8 -*-
import collections
import inspect
import numbers
import re

import numpy as np


def _get_docstring(func):
    # Drop the output parameter from the docstring.
    split_doc_params = lambda s: re.subn(                         # noqa: E731
        "(    [A-Za-z]+ : )", "\0\\1", s)[0].split("\0")
    drop_doc_param = lambda s: not s.startswith("    output : ")  # noqa: E731
    func_doc = "" if func.__doc__ is None else func.__doc__
    cleaned_docstring = "".join([
        l for l in split_doc_params(func_doc) if drop_doc_param(l)  # noqa: E741, E501
    ])
    cleaned_docstring = cleaned_docstring.replace('input', 'image')
    cleaned_docstring = cleaned_docstring.replace('labels', 'label_image')
    cleaned_docstring = cleaned_docstring.split('Examples')[0].strip()

    docstring = """
    Wrapped copy of "{mod_name}.{func_name}"


    Excludes the output parameter as it would not work with Dask arrays.


    Original docstring:

    {doc}
    """.format(
        mod_name=inspect.getmodule(func).__name__,
        func_name=func.__name__,
        doc=cleaned_docstring,
    )

    return docstring


def _update_wrapper(func):
    def _updater(wrapper):
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = _get_docstring(func)
        return wrapper

    return _updater


def _get_depth_boundary(ndim, depth, boundary=None):
    strlike = (bytes, str)

    if not isinstance(ndim, numbers.Integral):
        raise TypeError("Expected integer value for `ndim`.")
    if ndim <= 0:
        raise ValueError("Expected positive value for `ndim`.")

    if isinstance(depth, numbers.Number):
        depth = ndim * (depth,)
    if not isinstance(depth, collections.abc.Sized):
        raise TypeError("Unexpected type for `depth`.")
    if len(depth) != ndim:
        raise ValueError("Expected `depth` to have a length equal to `ndim`.")
    if isinstance(depth, collections.abc.Sequence):
        depth = dict(zip(range(ndim), depth))
    if not isinstance(depth, collections.abc.Mapping):
        raise TypeError("Unexpected type for `depth`.")

    if not all(map(lambda d: isinstance(d, numbers.Integral), depth.values())):
        raise TypeError("Expected integer values for `depth`.")
    if not all(map(lambda d: d >= 0, depth.values())):
        raise ValueError("Expected positive semidefinite values for `depth`.")

    depth = dict([(a, int(d)) for a, d in depth.items()])

    if (boundary is None) or isinstance(boundary, strlike):
        boundary = ndim * (boundary,)
    if not isinstance(boundary, collections.abc.Sized):
        raise TypeError("Unexpected type for `boundary`.")
    if len(boundary) != ndim:
        raise ValueError(
            "Expected `boundary` to have a length equal to `ndim`."
        )
    if isinstance(boundary, collections.abc.Sequence):
        boundary = dict(zip(range(ndim), boundary))
    if not isinstance(boundary, collections.abc.Mapping):
        raise TypeError("Unexpected type for `boundary`.")

    type_check = lambda b: (b is None) or isinstance(b, strlike)  # noqa: E731
    if not all(map(type_check, boundary.values())):
        raise TypeError("Expected string-like values for `boundary`.")

    return depth, boundary


def _get_size(ndim, size):
    if not isinstance(ndim, numbers.Integral):
        raise TypeError("The ndim must be of integral type.")

    if isinstance(size, numbers.Number):
        size = ndim * (size,)
    size = np.array(size)

    if size.ndim != 1:
        raise RuntimeError("The size must have only one dimension.")
    if len(size) != ndim:
        raise RuntimeError(
            "The size must have a length equal to the number of dimensions."
        )
    if not issubclass(size.dtype.type, numbers.Integral):
        raise TypeError("The size must be of integral type.")

    size = tuple(size)

    return size


def _get_origin(size, origin=0):
    size = np.array(size)
    ndim = len(size)

    if isinstance(origin, numbers.Number):
        origin = ndim * (origin,)

    origin = np.array(origin)

    if not issubclass(origin.dtype.type, numbers.Integral):
        raise TypeError("The origin must be of integral type.")

    # Validate dimensions.
    if origin.ndim != 1:
        raise RuntimeError("The origin must have only one dimension.")
    if len(origin) != ndim:
        raise RuntimeError(
            "The origin must have the same length as the number of dimensions"
            " as the array being filtered."
        )

    # Validate origin is bounded.
    if not (origin < ((size + 1) // 2)).all():
        raise ValueError("The origin must be within the footprint.")

    origin = tuple(origin)

    return origin


def _get_depth(size, origin=0):
    origin = np.array(_get_origin(size, origin))
    size = np.array(size)

    half_size = size // 2
    depth = half_size + abs(origin)

    depth = tuple(depth)

    return depth


def _get_footprint(ndim, size=None, footprint=None):
    # Verify that we only got size or footprint.
    if size is None and footprint is None:
        raise RuntimeError("Must provide either size or footprint.")
    if size is not None and footprint is not None:
        raise RuntimeError("Provide either size or footprint, but not both.")

    # Get a footprint based on the size.
    if size is not None:
        size = _get_size(ndim, size)
        footprint = np.ones(size, dtype=bool)

    # Validate the footprint.
    if footprint.ndim != ndim:
        raise RuntimeError(
            "The footprint must have the same number of dimensions as"
            " the array being filtered."
        )
    if footprint.size == 0:
        raise RuntimeError("The footprint must have only non-zero dimensions.")

    # Convert to Boolean.
    footprint = (footprint != 0)

    return footprint



# -*- coding: utf-8 -*-
import numbers

import pytest

import dask.array as da

import dask_image.ndfourier._utils


@pytest.mark.parametrize(
    "a, s, n, axis", [
        (da.ones((3, 4), chunks=(3, 4)), da.ones((2,), chunks=(2,)), -1, -1),
        (da.ones((3, 4), dtype='i8', chunks=(3, 4)),
         da.ones((2,), dtype='i8', chunks=(2,)), -1, -1),
    ]
)
def test_norm_args(a, s, n, axis):
    a2, s2, n2, axis2 = dask_image.ndfourier._utils._norm_args(
        a, s, n=n, axis=axis
    )

    assert isinstance(a2, da.Array)
    assert isinstance(s2, da.Array)

    assert issubclass(a2.real.dtype.type, numbers.Real)
    assert issubclass(s2.dtype.type, numbers.Real)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import inspect

import pytest
import numpy as np

from dask_image.ndfilters import _utils


def test__get_docstring():
    f = lambda: 0  # noqa: E731

    result = _utils._get_docstring(f)

    expected = """
    Wrapped copy of "{mod_name}.{func_name}"


    Excludes the output parameter as it would not work with Dask arrays.


    Original docstring:

    {doc}
    """.format(
        mod_name=inspect.getmodule(f).__name__,
        func_name=f.__name__,
        doc="",
    )

    assert result == expected


def test__update_wrapper():
    f = lambda: 0  # noqa: E731

    @_utils._update_wrapper(f)
    def g():
        return f()

    assert f.__name__ == g.__name__

    expected = """
    Wrapped copy of "{mod_name}.{func_name}"


    Excludes the output parameter as it would not work with Dask arrays.


    Original docstring:

    {doc}
    """.format(
        mod_name=inspect.getmodule(g).__name__,
        func_name=g.__name__,
        doc="",
    )

    assert g.__doc__ == expected


@pytest.mark.parametrize(
    "err_type, ndim, depth, boundary",
    [
        (TypeError, lambda: 0, 1, None),
        (TypeError, 1.0, 1, None),
        (ValueError, -1, 1, None),
        (TypeError, 1, lambda: 0, None),
        (TypeError, 1, 1.0, None),
        (ValueError, 1, -1, None),
        (ValueError, 1, (1, 1), None),
        (ValueError, 1, {0: 1, 1: 1}, None),
        (TypeError, 1, {1}, None),
        (TypeError, 1, 1, 1),
        (ValueError, 1, 1, (None, None)),
        (ValueError, 1, 1, {0: None, 1: None}),
        (TypeError, 1, 1, (1,)),
        (TypeError, 1, 1, {1}),
    ]
)
def test_errs__get_depth_boundary(err_type, ndim, depth, boundary):
    with pytest.raises(err_type):
        _utils._get_depth_boundary(ndim, depth, boundary)


@pytest.mark.parametrize(
    "err_type, ndim, size",
    [
        (TypeError, 1.0, 1),
        (RuntimeError, 1, [[1]]),
        (TypeError, 1, 1.0),
        (TypeError, 1, [1.0]),
        (RuntimeError, 1, [1, 1]),
    ]
)
def test_errs__get_size(err_type, ndim, size):
    with pytest.raises(err_type):
        _utils._get_size(ndim, size)


@pytest.mark.parametrize(
    "err_type, size, origin",
    [
        (TypeError, [1], 1.0),
        (TypeError, [1], [1.0]),
        (RuntimeError, [1], [[1]]),
        (RuntimeError, [1], [1, 1]),
        (ValueError, [1], [2]),
    ]
)
def test_errs__get_origin(err_type, size, origin):
    with pytest.raises(err_type):
        _utils._get_origin(size, origin)


@pytest.mark.parametrize(
    "err_type, ndim, size, footprint",
    [
        (RuntimeError, 1, None, None),
        (RuntimeError, 1, [2], np.ones((2,), dtype=bool)),
        (RuntimeError, 1, None, np.ones((1, 2), dtype=bool)),
        (RuntimeError, 1, None, np.ones([0], dtype=bool)),
    ]
)
def test_errs__get_footprint(err_type, ndim, size, footprint):
    with pytest.raises(err_type):
        _utils._get_footprint(ndim, size=size, footprint=footprint)


@pytest.mark.parametrize(
    "expected, ndim, depth, boundary",
    [
        (({0: 0}, {0: "none"}), 1, 0, "none"),
        (({0: 0}, {0: "reflect"}), 1, 0, "reflect"),
        (({0: 0}, {0: "periodic"}), 1, 0, "periodic"),
        (({0: 1}, {0: "none"}), 1, 1, "none"),
    ]
)
def test__get_depth_boundary(expected, ndim, depth, boundary):
    assert expected == _utils._get_depth_boundary(ndim, depth, boundary)


@pytest.mark.parametrize(
    "expected, ndim, size",
    [
        ((1,), 1, 1),
        ((3, 3), 2, 3),
        ((2, 4), 2, (2, 4)),
    ]
)
def test__get_size(expected, ndim, size):
    assert expected == _utils._get_size(ndim, size)


@pytest.mark.parametrize(
    "expected, size, origin",
    [
        ((0,), (1,), 0),
        ((1,), (3,), 1),
        ((1, 2), (3, 5), (1, 2)),
    ]
)
def test__get_origin(expected, size, origin):
    assert expected == _utils._get_origin(size, origin)


@pytest.mark.parametrize(
    "expected, size, origin",
    [
        ((0,), (1,), 0),
        ((1,), (3,), 0),
        ((2,), (3,), 1),
        ((2, 4), (3, 5), (1, 2)),
    ]
)
def test__get_depth(expected, size, origin):
    assert expected == _utils._get_depth(size, origin)


@pytest.mark.parametrize(
    "expected, ndim, size, footprint",
    [
        (np.ones((2,), dtype=bool), 1, 2, None),
        (np.ones((2,), dtype=bool), 1, None, np.ones((2,), dtype=bool)),
    ]
)
def test__get_footprint(expected, ndim, size, footprint):
    assert (expected == _utils._get_footprint(ndim, size, footprint)).all()



# -*- coding: utf-8 -*-

import numpy as np
import scipy.ndimage

from ._dispatcher import Dispatcher

__all__ = [
    "dispatch_binary_dilation",
    "dispatch_binary_erosion",
    "dispatch_binary_structure",
]

dispatch_binary_dilation = Dispatcher(name="dispatch_binary_dilation")
dispatch_binary_erosion = Dispatcher(name="dispatch_binary_erosion")
dispatch_binary_structure = Dispatcher(name='dispatch_binary_structure')


# ================== binary_dilation ==================
@dispatch_binary_dilation.register(np.ndarray)
def numpy_binary_dilation(*args, **kwargs):
    return scipy.ndimage.binary_dilation


@dispatch_binary_dilation.register_lazy("cupy")
def register_cupy_binary_dilation():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_binary_dilation.register(cupy.ndarray)
    def cupy_binary_dilation(*args, **kwargs):
        return cupyx.scipy.ndimage.binary_dilation


# ================== binary_erosion ==================
@dispatch_binary_erosion.register(np.ndarray)
def numpy_binary_erosion(*args, **kwargs):
    return scipy.ndimage.binary_erosion


@dispatch_binary_erosion.register_lazy("cupy")
def register_cupy_binary_erosion():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_binary_erosion.register(cupy.ndarray)
    def cupy_binary_erosion(*args, **kwargs):
        return cupyx.scipy.ndimage.binary_erosion


# ================== generate_binary_structure ==================
@dispatch_binary_structure.register(np.ndarray)
def numpy_binary_structure(*args, **kwargs):
    return scipy.ndimage.generate_binary_structure


@dispatch_binary_structure.register_lazy("cupy")
def register_cupy_binary_structure():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_binary_structure.register(cupy.ndarray)
    def cupy_binary_structure(*args, **kwargs):
        return cupyx.scipy.ndimage.generate_binary_structure



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import numpy as np
import scipy.ndimage

import dask.array as da

import dask_image.ndfilters


@pytest.mark.parametrize(
    "err_type, sigma, truncate",
    [
        (RuntimeError, [[1.0]], 4.0),
        (RuntimeError, [1.0], 4.0),
        (TypeError, 1.0 + 0.0j, 4.0),
        (TypeError, 1.0, 4.0 + 0.0j),
    ]
)
@pytest.mark.parametrize(
    "da_func",
    [
        dask_image.ndfilters.gaussian,
        dask_image.ndfilters.gaussian_filter,
        dask_image.ndfilters.gaussian_gradient_magnitude,
        dask_image.ndfilters.gaussian_laplace,
    ]
)
def test_gaussian_filters_params(da_func, err_type, sigma, truncate):
    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    with pytest.raises(err_type):
        da_func(d, sigma, truncate=truncate)


@pytest.mark.parametrize(
    "sigma, truncate",
    [
        (0.0, 0.0),
        (0.0, 1.0),
        (0.0, 4.0),
        (1.0, 0.0),
    ]
)
@pytest.mark.parametrize(
    "order", [0, 1, 2, 3]
)
@pytest.mark.parametrize(
    "sp_func, da_func",
    [
        (scipy.ndimage.gaussian_filter, dask_image.ndfilters.gaussian),
        (scipy.ndimage.gaussian_filter, dask_image.ndfilters.gaussian_filter),
    ]
)
def test_gaussian_filters_identity(sp_func, da_func, order, sigma, truncate):
    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    if order % 2 == 1 and sigma != 0 and truncate == 0:
        pytest.skip(
            "SciPy zeros the result of a Gaussian filter with odd derivatives"
            " when sigma is non-zero, truncate is zero, and derivative is odd."
            "\n\nxref: https://github.com/scipy/scipy/issues/7364"
        )

    da.utils.assert_eq(
        d, da_func(d, sigma, order, truncate=truncate)
    )

    da.utils.assert_eq(
        sp_func(a, sigma, order, truncate=truncate),
        da_func(d, sigma, order, truncate=truncate)
    )


@pytest.mark.parametrize(
    "da_func",
    [
        dask_image.ndfilters.gaussian,
        dask_image.ndfilters.gaussian_filter,
        dask_image.ndfilters.gaussian_gradient_magnitude,
        dask_image.ndfilters.gaussian_laplace,
    ]
)
def test_gaussian_filter_shape_type(da_func):
    sigma = 1.0
    truncate = 4.0

    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    assert all([(type(s) is int) for s in d.shape])

    d2 = da_func(d, sigma=sigma, truncate=truncate)

    assert all([(type(s) is int) for s in d2.shape])


@pytest.mark.parametrize(
    "da_func",
    [
        dask_image.ndfilters.gaussian,
        dask_image.ndfilters.gaussian_filter,
        dask_image.ndfilters.gaussian_gradient_magnitude,
        dask_image.ndfilters.gaussian_laplace,
    ]
)
def test_gaussian_filter_comprehensions(da_func):
    da_wfunc = lambda arr: da_func(arr, 1.0, truncate=4.0)  # noqa: E731

    np.random.seed(0)

    a = np.random.random((3, 12, 14))
    d = da.from_array(a, chunks=(3, 6, 7))

    l2s = [da_wfunc(d[i]) for i in range(len(d))]
    l2c = [da_wfunc(d[i])[None] for i in range(len(d))]

    da.utils.assert_eq(np.stack(l2s), da.stack(l2s))
    da.utils.assert_eq(np.concatenate(l2c), da.concatenate(l2c))


@pytest.mark.parametrize(
    "sigma, truncate",
    [
        (1.0, 2.0),
        (1.0, 4.0),
        (2.0, 2.0),
        (2.0, 4.0),
        ((1.0, 2.0), 4.0),
    ]
)
@pytest.mark.parametrize(
    "sp_func, da_func",
    [
        (scipy.ndimage.gaussian_filter, dask_image.ndfilters.gaussian),
        (scipy.ndimage.gaussian_filter, dask_image.ndfilters.gaussian_filter),
        (scipy.ndimage.gaussian_gradient_magnitude,
         dask_image.ndfilters.gaussian_gradient_magnitude),
        (scipy.ndimage.gaussian_laplace,
         dask_image.ndfilters.gaussian_laplace),
    ]
)
def test_gaussian_filters_compare(sp_func, da_func, sigma, truncate):
    s = (100, 110)
    a = np.arange(float(np.prod(s))).reshape(s)
    d = da.from_array(a, chunks=(50, 55))

    da.utils.assert_eq(
        sp_func(a, sigma, truncate=truncate),
        da_func(d, sigma, truncate=truncate)
    )


@pytest.mark.parametrize(
    "sigma, truncate",
    [
        (0.0, 0.0),
        (1.0, 0.0),
        (0.0, 1.0),
        (1.0, 2.0),
        (1.0, 4.0),
        (2.0, 2.0),
        (2.0, 4.0),
        ((1.0, 2.0), 4.0),
    ]
)
@pytest.mark.parametrize(
    "order", [
        0,
        1,
        2,
        3,
        (0, 1),
        (2, 3),
    ]
)
@pytest.mark.parametrize(
    "sp_func, da_func",
    [
        (scipy.ndimage.gaussian_filter, dask_image.ndfilters.gaussian),
        (scipy.ndimage.gaussian_filter, dask_image.ndfilters.gaussian_filter),
    ]
)
def test_gaussian_derivative_filters_compare(sp_func, da_func,
                                             order, sigma, truncate):
    s = (100, 110)
    a = np.arange(float(np.prod(s))).reshape(s)
    d = da.from_array(a, chunks=(50, 55))

    da.utils.assert_eq(
        sp_func(a, sigma, order, truncate=truncate),
        da_func(d, sigma, order, truncate=truncate)
    )



# -*- coding: utf-8 -*-
import warnings

import dask
import dask.array as da
import numpy as np


def _norm_input_labels_index(image, label_image=None, index=None):
    """
    Normalize arguments to a standard form.
    """

    image = da.asarray(image)

    if label_image is None:
        label_image = da.ones(
            image.shape, dtype=int, chunks=image.chunks,
        )
        index = da.from_array(np.array(1, dtype=int))
    elif index is None:
        label_image = (label_image > 0).astype(int)
        index = da.from_array(np.array(1, dtype=int))

    label_image = da.asarray(label_image)
    index = da.asarray(index)

    if index.ndim > 1:
        warnings.warn(
            "Having index with dimensionality greater than 1 is undefined.",
            FutureWarning
        )

    if image.shape != label_image.shape:
        raise ValueError(
            "The image and label_image arrays must be the same shape."
        )

    return (image, label_image, index)


def _ravel_shape_indices_kernel(*args):
    args2 = tuple(
        a[i * (None,) + (slice(None),) + (len(args) - i - 1) * (None,)]
        for i, a in enumerate(args)
    )
    return sum(args2)


def _ravel_shape_indices(dimensions, dtype=int, chunks=None):
    """
    Gets the raveled indices shaped like input.
    """

    indices = [
        da.arange(
            0,
            np.prod(dimensions[i:], dtype=dtype),
            np.prod(dimensions[i + 1:], dtype=dtype),
            dtype=dtype,
            chunks=c
        )
        for i, c in enumerate(chunks)
    ]

    indices = da.blockwise(
        _ravel_shape_indices_kernel, tuple(range(len(indices))),
        *sum([(a, (i,)) for i, a in enumerate(indices)], tuple()),
        dtype=dtype
    )

    return indices


def _argmax(a, positions, shape, dtype):
    """
    Find original array position corresponding to the maximum.
    """

    result = np.empty((1,), dtype=dtype)

    pos_nd = np.unravel_index(positions[np.argmax(a)], shape)
    for i, pos_nd_i in enumerate(pos_nd):
        result["pos"][0, i] = pos_nd_i

    return result[0]


def _argmin(a, positions, shape, dtype):
    """
    Find original array position corresponding to the minimum.
    """

    result = np.empty((1,), dtype=dtype)

    pos_nd = np.unravel_index(positions[np.argmin(a)], shape)
    for i, pos_nd_i in enumerate(pos_nd):
        result["pos"][0, i] = pos_nd_i

    return result[0]


def _center_of_mass(a, positions, shape, dtype):
    """
    Find the center of mass for each ROI.
    """

    result = np.empty((1,), dtype=dtype)

    positions_nd = np.unravel_index(positions, shape)
    a_sum = np.sum(a)

    a_wt_i = np.empty(a.shape)
    for i, pos_nd_i in enumerate(positions_nd):
        a_wt_sum_i = np.multiply(a, pos_nd_i, out=a_wt_i).sum()
        result["com"][0, i] = a_wt_sum_i / a_sum

    return result[0]


def _extrema(a, positions, shape, dtype):
    """
    Find minimum and maximum as well as positions for both.
    """

    result = np.empty((1,), dtype=dtype)

    int_min_pos = np.argmin(a)
    int_max_pos = np.argmax(a)

    result["min_val"] = a[int_min_pos]
    result["max_val"] = a[int_max_pos]

    min_pos_nd = np.unravel_index(positions[int_min_pos], shape)
    max_pos_nd = np.unravel_index(positions[int_max_pos], shape)
    for i in range(len(shape)):
        result["min_pos"][0, i] = min_pos_nd[i]
        result["max_pos"][0, i] = max_pos_nd[i]

    return result[0]


def _histogram(image,
               min,
               max,
               bins):
    """
    Delayed wrapping of NumPy's histogram

    Also reformats the arguments.
    """

    return np.histogram(image, bins, (min, max))[0]


@dask.delayed
def _labeled_comprehension_delayed(func,
                                   out_dtype,
                                   default,
                                   a,
                                   positions=None):
    """
    Wrapped delayed labeled comprehension function

    Included in the module for pickling purposes. Also handle cases where
    computation should not occur.
    """

    result = np.empty((1,), dtype=out_dtype)

    if a.size:
        if positions is None:
            result[0] = func(a)
        else:
            result[0] = func(a, positions)
    else:
        result[0] = default[0]

    return result


def _labeled_comprehension_func(func,
                                out_dtype,
                                default,
                                a,
                                positions=None):
    """
    Wrapped labeled comprehension function

    Ensures the result is a proper Dask Array and the computation delayed.
    """

    return da.from_delayed(
        _labeled_comprehension_delayed(func, out_dtype, default, a, positions),
        (1,),
        out_dtype
    )



# -*- coding: utf-8 -*-

import numpy as np
import scipy.ndimage

from ..dispatch._dispatch_ndfilters import dispatch_generic_filter
from . import _utils

__all__ = [
    "generic_filter",
]


@_utils._update_wrapper(scipy.ndimage.generic_filter)
def generic_filter(image,
                   function,
                   size=None,
                   footprint=None,
                   mode='reflect',
                   cval=0.0,
                   origin=0,
                   extra_arguments=tuple(),
                   extra_keywords=dict()):
    footprint = _utils._get_footprint(image.ndim, size, footprint)
    origin = _utils._get_origin(footprint.shape, origin)
    depth = _utils._get_depth(footprint.shape, origin)
    depth, boundary = _utils._get_depth_boundary(footprint.ndim, depth, "none")

    if type(image._meta) == np.ndarray:
        kwargs = {"extra_arguments": extra_arguments,
                  "extra_keywords": extra_keywords}
    else:  # pragma: no cover
        # cupy generic_filter doesn't support extra_arguments or extra_keywords
        kwargs = {}

    result = image.map_overlap(
        dispatch_generic_filter(image),
        depth=depth,
        boundary=boundary,
        dtype=image.dtype,
        meta=image._meta,
        function=function,
        footprint=footprint,
        mode=mode,
        cval=cval,
        origin=origin,
        **kwargs
    )

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import numpy as np
import scipy.ndimage

import dask.array as da

import dask_image.ndfilters


@pytest.mark.parametrize(
    "da_func",
    [
        (dask_image.ndfilters.convolve),
        (dask_image.ndfilters.correlate),
    ]
)
@pytest.mark.parametrize(
    "err_type, weights, origin",
    [
        (ValueError, np.ones((1,)), 0),
        (ValueError, np.ones((1, 0)), 0),
        (RuntimeError, np.ones((1, 1)), (0,)),
        (RuntimeError, np.ones((1, 1)), [(0,)]),
        (ValueError, np.ones((1, 1)), 1),
        (TypeError, np.ones((1, 1)), 0.0),
        (TypeError, np.ones((1, 1)), (0.0, 0.0)),
        (TypeError, np.ones((1, 1)), 1+0j),
        (TypeError, np.ones((1, 1)), (0+0j, 1+0j)),
    ]
)
def test_convolutions_params(da_func,
                             err_type,
                             weights,
                             origin):
    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    with pytest.raises(err_type):
        da_func(d,
                weights,
                origin=origin)


@pytest.mark.parametrize(
    "da_func",
    [
        dask_image.ndfilters.convolve,
        dask_image.ndfilters.correlate,
    ]
)
def test_convolutions_shape_type(da_func):
    weights = np.ones((1, 1))

    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    assert all([(type(s) is int) for s in d.shape])

    d2 = da_func(d, weights)

    assert all([(type(s) is int) for s in d2.shape])


@pytest.mark.parametrize(
    "da_func",
    [
        dask_image.ndfilters.convolve,
        dask_image.ndfilters.correlate,
    ]
)
def test_convolutions_comprehensions(da_func):
    np.random.seed(0)

    a = np.random.random((3, 12, 14))
    d = da.from_array(a, chunks=(3, 6, 7))

    weights = np.ones((1, 1))

    l2s = [da_func(d[i], weights) for i in range(len(d))]
    l2c = [da_func(d[i], weights)[None] for i in range(len(d))]

    da.utils.assert_eq(np.stack(l2s), da.stack(l2s))
    da.utils.assert_eq(np.concatenate(l2c), da.concatenate(l2c))


@pytest.mark.parametrize(
    "sp_func, da_func",
    [
        (scipy.ndimage.convolve, dask_image.ndfilters.convolve),
        (scipy.ndimage.correlate, dask_image.ndfilters.correlate),
    ]
)
@pytest.mark.parametrize(
    "weights",
    [
        np.ones((1, 1)),
    ]
)
def test_convolutions_identity(sp_func,
                               da_func,
                               weights):
    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    da.utils.assert_eq(
        d, da_func(d, weights)
    )

    da.utils.assert_eq(
        sp_func(a, weights),
        da_func(d, weights)
    )


@pytest.mark.parametrize(
    "sp_func, da_func",
    [
        (scipy.ndimage.convolve, dask_image.ndfilters.convolve),
        (scipy.ndimage.correlate, dask_image.ndfilters.correlate),
    ]
)
@pytest.mark.parametrize(
    "weights, origin",
    [
        (np.ones((2, 2)), 0),
        (np.ones((2, 3)), 0),
        (np.ones((2, 3)), (0, 1)),
        (np.ones((2, 3)), (0, -1)),
        ((np.mgrid[-2: 2+1, -2: 2+1]**2).sum(axis=0) < 2.5**2, 0),
        ((np.mgrid[-2: 2+1, -2: 2+1]**2).sum(axis=0) < 2.5**2, (1, 2)),
        ((np.mgrid[-2: 2+1, -2: 2+1]**2).sum(axis=0) < 2.5**2, (-1, -2)),
        (np.ones((5, 5)), 0),
        (np.ones((7, 7)), 0),
        (np.ones((8, 8)), 0),
        (np.ones((10, 10)), 0),
        (np.ones((5, 5)), 2),
        (np.ones((5, 5)), -2),
    ]
)
def test_convolutions_compare(sp_func,
                              da_func,
                              weights,
                              origin):
    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    da.utils.assert_eq(
        sp_func(
            a, weights, origin=origin
        ),
        da_func(
            d, weights, origin=origin
        )
    )
@pytest.mark.parametrize(
    "sp_func, da_func",
    [
        (scipy.ndimage.convolve, dask_image.ndfilters.convolve),
        (scipy.ndimage.correlate, dask_image.ndfilters.correlate),
    ]
)
@pytest.mark.parametrize(
    "weights",
    [
     np.ones((1,5)),
     np.ones((5,1)),
    ]
)
@pytest.mark.parametrize(
    "mode",
    ["reflect","wrap","nearest","constant","mirror"])
def test_convolutions_modes(sp_func,
                            da_func,
                            weights,
                            mode):
    a = np.arange(140).reshape(10,14)
    d = da.from_array(a,chunks =(5, 7))
    
    da.utils.assert_eq(
        sp_func(
            a, weights, mode = mode
        ),
        da_func(
            d, weights, mode = mode
        )
    )


# -*- coding: utf-8 -*-

import numpy as np
from scipy import ndimage

from ._dispatcher import Dispatcher

__all__ = [
    "dispatch_affine_transform",
    "dispatch_asarray",
]

dispatch_affine_transform = Dispatcher(name="dispatch_affine_transform")


# ================== affine_transform ==================
@dispatch_affine_transform.register(np.ndarray)
def numpy_affine_transform(*args, **kwargs):
    return ndimage.affine_transform


@dispatch_affine_transform.register_lazy("cupy")
def register_cupy_affine_transform():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_affine_transform.register(cupy.ndarray)
    def cupy_affine_transform(*args, **kwargs):

        return cupyx.scipy.ndimage.affine_transform


dispatch_spline_filter = Dispatcher(name="dispatch_spline_filter")


# ================== spline_filter ==================
@dispatch_spline_filter.register(np.ndarray)
def numpy_spline_filter(*args, **kwargs):
    return ndimage.spline_filter


@dispatch_spline_filter.register_lazy("cupy")
def register_cupy_spline_filter():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_spline_filter.register(cupy.ndarray)
    def cupy_spline_filter(*args, **kwargs):

        return cupyx.scipy.ndimage.spline_filter


dispatch_spline_filter1d = Dispatcher(name="dispatch_spline_filter1d")


# ================== spline_filter1d ==================
@dispatch_spline_filter1d.register(np.ndarray)
def numpy_spline_filter1d(*args, **kwargs):
    return ndimage.spline_filter1d


@dispatch_spline_filter1d.register_lazy("cupy")
def register_cupy_spline_filter1d():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_spline_filter1d.register(cupy.ndarray)
    def cupy_spline_filter1d(*args, **kwargs):

        return cupyx.scipy.ndimage.spline_filter1d


dispatch_asarray = Dispatcher(name="dispatch_asarray")


# ===================== asarray ========================
@dispatch_asarray.register(np.ndarray)
def numpy_asarray(*args, **kwargs):
    return np.asarray


@dispatch_asarray.register_lazy("cupy")
def register_cupy_asarray():
    import cupy

    @dispatch_asarray.register(cupy.ndarray)
    def cupy_asarray(*args, **kwargs):

        return cupy.asarray



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import numpy as np
import dask.array as da

from dask_image.ndmorph import _utils


@pytest.mark.parametrize(
    "err_type, input, structure",
    [
        (
            RuntimeError,
            da.ones([1, 2], dtype=bool, chunks=(1, 2,)),
            da.arange(2, dtype=bool, chunks=(2,))
        ),
        (
            TypeError,
            da.arange(2, dtype=bool, chunks=(2,)),
            2.0
        ),
    ]
)
def test_errs__get_structure(err_type, input, structure):
    with pytest.raises(err_type):
        _utils._get_structure(input, structure)


@pytest.mark.parametrize(
    "err_type, iterations",
    [
        (TypeError, 0.0),
        (NotImplementedError, 0),
    ]
)
def test_errs__get_iterations(err_type, iterations):
    with pytest.raises(err_type):
        _utils._get_iterations(iterations)


@pytest.mark.parametrize(
    "err_type, input, mask",
    [
        (
            RuntimeError,
            da.arange(2, dtype=bool, chunks=(2,)),
            da.arange(1, dtype=bool, chunks=(2,))
        ),
        (
            TypeError,
            da.arange(2, dtype=bool, chunks=(2,)),
            2.0
        ),
    ]
)
def test_errs__get_mask(err_type, input, mask):
    with pytest.raises(err_type):
        _utils._get_mask(input, mask)


@pytest.mark.parametrize(
    "err_type, border_value",
    [
        (TypeError, 0.0),
        (TypeError, 1.0),
    ]
)
def test_errs__get_border_value(err_type, border_value):
    with pytest.raises(err_type):
        _utils._get_border_value(border_value)


@pytest.mark.parametrize(
    "err_type, brute_force",
    [
        (NotImplementedError, True),
        (TypeError, 1),
    ]
)
def test_errs__get_brute_force(err_type, brute_force):
    with pytest.raises(err_type):
        _utils._get_brute_force(brute_force)


@pytest.mark.parametrize(
    "expected, input, structure",
    [
        (
            np.array([1, 1, 1], dtype=bool),
            (da.arange(10, chunks=(10,)) % 2).astype(bool),
            None
        ),
        (
            np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]], dtype=bool),
            (da.arange(100, chunks=10).reshape(10, 10) % 2).astype(bool),  # noqa: E501
            None
        ),
        (
            np.array([1, 1, 1], dtype=bool),
            (da.arange(10, chunks=(10,)) % 2).astype(bool),
            np.array([1, 1, 1], dtype=int)
        ),
        (
            np.array([1, 1, 1], dtype=bool),
            (da.arange(10, chunks=(10,)) % 2).astype(bool),
            np.array([1, 1, 1], dtype=bool)
        ),
    ]
)
def test__get_structure(expected, input, structure):
    result = _utils._get_structure(input, structure)

    assert expected.dtype.type == result.dtype.type
    assert np.array((expected == result).all())[()]


@pytest.mark.parametrize(
    "expected, iterations",
    [
        (1, 1),
        (4, 4),
    ]
)
def test__get_iterations(expected, iterations):
    assert expected == _utils._get_iterations(iterations)


@pytest.mark.parametrize(
    "expected, a",
    [
        (np.bool8, False),
        (np.int_, 2),
        (np.float64, 3.1),
        (np.complex128, 1 + 2j),
        (np.int16, np.int16(6)),
        (np.uint32, np.arange(3, dtype=np.uint32)),
    ]
)
def test__get_dtype(expected, a):
    assert expected == _utils._get_dtype(a)


@pytest.mark.parametrize(
    "expected, input, mask",
    [
        (True, da.arange(2, dtype=bool, chunks=(2,)), None),
        (True, da.arange(2, dtype=bool, chunks=(2,)), True),
        (False, da.arange(2, dtype=bool, chunks=(2,)), False),
        (
            True,
            da.arange(2, dtype=bool, chunks=(2,)),
            np.bool8(True)
        ),
        (
            False,
            da.arange(2, dtype=bool, chunks=(2,)),
            np.bool8(False)
        ),
        (
            np.arange(2, dtype=bool),
            da.arange(2, dtype=bool, chunks=(2,)),
            np.arange(2, dtype=bool)
        ),
        (
            da.arange(2, dtype=bool, chunks=(2,)),
            da.arange(2, dtype=bool, chunks=(2,)),
            da.arange(2, dtype=int, chunks=(2,))
        ),
    ]
)
def test__get_mask(expected, input, mask):
    result = _utils._get_mask(input, mask)

    assert type(expected) == type(result)

    if isinstance(expected, (np.ndarray, da.Array)):
        assert np.array((expected == result).all())[()]
    else:
        assert expected == result


@pytest.mark.parametrize(
    "expected, border_value",
    [
        (False, False),
        (True, True),
        (False, 0),
        (True, 1),
        (True, 5),
        (True, -2),
    ]
)
def test__get_border_value(expected, border_value):
    assert expected == _utils._get_border_value(border_value)


@pytest.mark.parametrize(
    "expected, brute_force",
    [
        (False, False),
    ]
)
def test__get_brute_force(expected, brute_force):
    assert expected == _utils._get_brute_force(brute_force)



from dask_image.ndmeasure._utils import _labeled_comprehension_delayed
import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd
import pytest

import dask_image.ndmeasure


@pytest.fixture
def label_image():
    """Return small label image for tests.

    dask.array<array, shape=(5, 10), dtype=int64, chunksize=(5, 5), chunktype=numpy.ndarray>

    array([[   0,   0,   0,   0,   0,   0,   0, 333, 333, 333],
            [111, 111,   0,   0,   0,   0,   0, 333, 333, 333],
            [111, 111,   0,   0,   0,   0,   0,   0,   0,   0],
            [  0,   0,   0, 222, 222, 222, 222, 222, 222,   0],
            [  0,   0,   0,   0,   0,   0,   0,   0,   0,   0]])

    """
    label_image = np.zeros((5, 10)).astype(int)
    label_image[1:3,0:2] = 111
    label_image[3,3:-2] = 222
    label_image[0:2,-3:] = 333
    label_image = da.from_array(label_image, chunks=(5, 5))
    return label_image


@pytest.fixture
def label_image_with_empty_chunk():
    """Return small label image with an empty chunk for tests.

    dask.array<array, shape=(6, 6), dtype=int64, chunksize=(3, 3), chunktype=numpy.ndarray>

    array([[   0,   0,   0,   0,   0,   0],
            [111, 111,   0,   0,   0,   0],
            [111, 111,   0,   0,   0,   0],
            [  0,   0,   0,   0,   0,   0],
            [  0,   0,   0, 222, 222, 222],
            [  0,   0,   0,   0,   0,   0]])
    """
    label_image = np.zeros((6, 6)).astype(int)
    label_image[1:3,0:2] = 111
    label_image[4,3:] = 222
    label_image = da.from_array(label_image, chunks=(3, 3))
    return label_image


def test_find_objects_err(label_image):
    label_image = label_image.astype(float)
    with pytest.raises(ValueError):
        dask_image.ndmeasure.find_objects(label_image)


def test_find_objects(label_image):
    result = dask_image.ndmeasure.find_objects(label_image)
    assert isinstance(result, dd.DataFrame)
    computed_result = result.compute()
    assert isinstance(computed_result, pd.DataFrame)
    expected = pd.DataFrame.from_dict(
        {0: {111: slice(1, 3), 222: slice(3, 4), 333: slice(0, 2)},
        1: {111: slice(0, 2), 222: slice(3, 8), 333: slice(7, 10)}}
    )
    assert computed_result.equals(expected)


def test_3d_find_objects(label_image):
    label_image = da.stack([label_image, label_image], axis=2)
    result = dask_image.ndmeasure.find_objects(label_image)
    assert isinstance(result, dd.DataFrame)
    computed_result = result.compute()
    assert isinstance(computed_result, pd.DataFrame)
    expected = pd.DataFrame.from_dict(
        {0: {111: slice(1, 3), 222: slice(3, 4), 333: slice(0, 2)},
        1: {111: slice(0, 2), 222: slice(3, 8), 333: slice(7, 10)},
        2: {111: slice(0, 2), 222: slice(0, 2), 333: slice(0, 2)}}
    )
    assert computed_result.equals(expected)


def test_find_objects_with_empty_chunks(label_image_with_empty_chunk):
    result = dask_image.ndmeasure.find_objects(label_image_with_empty_chunk)
    assert isinstance(result, dd.DataFrame)
    computed_result = result.compute()
    assert isinstance(computed_result, pd.DataFrame)
    expected = pd.DataFrame.from_dict(
        {0: {111: slice(1, 3, None), 222: slice(4, 5, None)},
        1: {111: slice(0, 2, None), 222: slice(3, 6, None)}}
    )
    assert computed_result.equals(expected)



# -*- coding: utf-8 -*-


import numbers

import dask.array as da
import numpy as np

from ..dispatch._dispatch_ndmorph import dispatch_binary_structure
from ..ndfilters._utils import (_get_depth, _get_depth_boundary, _get_origin,
                                _update_wrapper)

_update_wrapper = _update_wrapper
_get_depth_boundary = _get_depth_boundary
_get_origin = _get_origin
_get_depth = _get_depth


def _get_structure(image, structure):
    # Create square connectivity as default
    if structure is None:
        generate_binary_structure = dispatch_binary_structure(image)
        structure = generate_binary_structure(image.ndim, 1)
    elif hasattr(structure, 'ndim'):
        if structure.ndim != image.ndim:
            raise RuntimeError(
                "`structure` must have the same rank as `image`."
            )
        if not issubclass(structure.dtype.type, np.bool8):
            structure = (structure != 0)
    else:
        raise TypeError("`structure` must be an array.")

    return structure


def _get_iterations(iterations):
    if not isinstance(iterations, numbers.Integral):
        raise TypeError("`iterations` must be of integral type.")
    if iterations < 1:
        raise NotImplementedError(
            "`iterations` must be equal to 1 or greater not less."
        )

    return iterations


def _get_dtype(a):
    # Get the dtype of a value or an array.
    # Even handle non-NumPy types.
    return getattr(a, "dtype", np.dtype(type(a)))


def _get_mask(image, mask):
    if mask is None:
        mask = True

    mask_type = _get_dtype(mask).type
    if isinstance(mask, (np.ndarray, da.Array)):
        if mask.shape != image.shape:
            raise RuntimeError("`mask` must have the same shape as `image`.")
        if not issubclass(mask_type, np.bool8):
            mask = (mask != 0)
    elif issubclass(mask_type, np.bool8):
        mask = bool(mask)
    else:
        raise TypeError("`mask` must be a Boolean or an array.")

    return mask


def _get_border_value(border_value):
    if not isinstance(border_value, numbers.Integral):
        raise TypeError("`border_value` must be of integral type.")

    border_value = (border_value != 0)

    return border_value


def _get_brute_force(brute_force):
    if brute_force is not False:
        if brute_force is True:
            raise NotImplementedError(
                "`brute_force` other than `False` is not yet supported."
            )
        else:
            raise TypeError(
                "`brute_force` must be `bool`."
            )

    return brute_force



# -*- coding: utf-8 -*-


import scipy.ndimage

from ..dispatch._dispatch_ndfilters import dispatch_uniform_filter
from . import _utils
from ._gaussian import gaussian_filter

__all__ = [
    "uniform_filter",
]

gaussian_filter = gaussian_filter


@_utils._update_wrapper(scipy.ndimage.uniform_filter)
def uniform_filter(image,
                   size=3,
                   mode='reflect',
                   cval=0.0,
                   origin=0):
    size = _utils._get_size(image.ndim, size)
    depth = _utils._get_depth(size, origin)

    depth, boundary = _utils._get_depth_boundary(image.ndim, depth, "none")

    result = image.map_overlap(
        dispatch_uniform_filter(image),
        depth=depth,
        boundary=boundary,
        dtype=image.dtype,
        meta=image._meta,
        size=size,
        mode=mode,
        cval=cval,
        origin=origin
    )

    return result



# -*- coding: utf-8 -*-

import functools
import math
from itertools import product
import warnings

import dask.array as da
import numpy as np
from dask.base import tokenize
from dask.highlevelgraph import HighLevelGraph
import scipy
from scipy.ndimage import affine_transform as ndimage_affine_transform

from ..dispatch._dispatch_ndinterp import (
    dispatch_affine_transform,
    dispatch_asarray,
    dispatch_spline_filter,
    dispatch_spline_filter1d,
)
from ..ndfilters._utils import _get_depth_boundary

from ..dispatch._dispatch_ndinterp import (dispatch_affine_transform,
                                           dispatch_asarray)

__all__ = [
    "affine_transform",
]


def affine_transform(
        image,
        matrix,
        offset=0.0,
        output_shape=None,
        order=1,
        output_chunks=None,
        **kwargs
):
    """Apply an affine transform using Dask. For every
    output chunk, only the slice containing the relevant part
    of the image is processed. Chunkwise processing is performed
    either using `ndimage.affine_transform` or
    `cupyx.scipy.ndimage.affine_transform`, depending on the input type.

    Notes
    -----
        Differences to `ndimage.affine_transformation`:
        - currently, prefiltering is not supported
          (affecting the output in case of interpolation `order > 1`)
        - default order is 1
        - modes 'reflect', 'mirror' and 'wrap' are not supported

        Arguments equal to `ndimage.affine_transformation`,
        except for `output_chunks`.

    Parameters
    ----------
    image : array_like (Numpy Array, Cupy Array, Dask Array...)
        The image array.
    matrix : array (ndim,), (ndim, ndim), (ndim, ndim+1) or (ndim+1, ndim+1)
        Transformation matrix.
    offset : float or sequence, optional
        The offset into the array where the transform is applied. If a float,
        `offset` is the same for each axis. If a sequence, `offset` should
        contain one value for each axis.
    output_shape : tuple of ints, optional
        The shape of the array to be returned.
    order : int, optional
        The order of the spline interpolation. Note that for order>1
        scipy's affine_transform applies prefiltering, which is not
        yet supported and skipped in this implementation.
    output_chunks : tuple of ints, optional
        The shape of the chunks of the output Dask Array.

    Returns
    -------
    affine_transform : Dask Array
        A dask array representing the transformed output

    """

    if not type(image) == da.core.Array:
        image = da.from_array(image)

    if output_shape is None:
        output_shape = image.shape

    if output_chunks is None:
        output_chunks = image.shape

    # Perform test run to ensure parameter validity.
    ndimage_affine_transform(np.zeros([0] * image.ndim),
                             matrix,
                             offset)

    # Make sure parameters contained in matrix and offset
    # are not overlapping, i.e. that the offset is valid as
    # it needs to be modified for each chunk.
    # Further parameter checks are performed directly by
    # `ndimage.affine_transform`.

    matrix = np.asarray(matrix)
    offset = np.asarray(offset).squeeze()

    # these lines were copied and adapted from `ndimage.affine_transform`
    if (matrix.ndim == 2 and matrix.shape[1] == image.ndim + 1 and
            (matrix.shape[0] in [image.ndim, image.ndim + 1])):

        # assume input is homogeneous coordinate transformation matrix
        offset = matrix[:image.ndim, image.ndim]
        matrix = matrix[:image.ndim, :image.ndim]

    cval = kwargs.pop('cval', 0)
    mode = kwargs.pop('mode', 'constant')
    prefilter = kwargs.pop('prefilter', False)

    supported_modes = ['constant', 'nearest']
    if scipy.__version__ > np.lib.NumpyVersion('1.6.0'):
        supported_modes += ['grid-constant']
    if mode in ['wrap', 'reflect', 'mirror', 'grid-mirror', 'grid-wrap']:
        raise NotImplementedError(
            f"Mode {mode} is not currently supported. It must be one of "
            f"{supported_modes}.")

    # process kwargs
    if prefilter and order > 1:
        # prefilter is not yet supported for all modes
        if mode in ['nearest', 'grid-constant']:
            raise NotImplementedError(
                f"order > 1 with mode='{mode}' is not supported. Currently "
                f"prefilter is only supported with mode='constant'."
            )
        image = spline_filter(image, order, output=np.float64,
                              mode=mode)

    n = image.ndim
    image_shape = image.shape

    # calculate output array properties
    normalized_chunks = da.core.normalize_chunks(output_chunks,
                                                 tuple(output_shape))
    block_indices = product(*(range(len(bds)) for bds in normalized_chunks))
    block_offsets = [np.cumsum((0,) + bds[:-1]) for bds in normalized_chunks]

    # use dispatching mechanism to determine backend
    affine_transform_method = dispatch_affine_transform(image)
    asarray_method = dispatch_asarray(image)

    # construct dask graph for output array
    # using unique and deterministic identifier
    output_name = 'affine_transform-' + tokenize(image, matrix, offset,
                                                 output_shape, output_chunks,
                                                 kwargs)
    output_layer = {}
    rel_images = []
    for ib, block_ind in enumerate(block_indices):

        out_chunk_shape = [normalized_chunks[dim][block_ind[dim]]
                           for dim in range(n)]
        out_chunk_offset = [block_offsets[dim][block_ind[dim]]
                            for dim in range(n)]

        out_chunk_edges = np.array([i for i in np.ndindex(tuple([2] * n))])\
            * np.array(out_chunk_shape) + np.array(out_chunk_offset)

        # map output chunk edges onto input image coordinates
        # to define the input region relevant for the current chunk
        if matrix.ndim == 1 and len(matrix) == image.ndim:
            rel_image_edges = matrix * out_chunk_edges + offset
        else:
            rel_image_edges = np.dot(matrix, out_chunk_edges.T).T + offset

        rel_image_i = np.min(rel_image_edges, 0)
        rel_image_f = np.max(rel_image_edges, 0)

        # Calculate edge coordinates required for the footprint of the
        # spline kernel according to
        # https://github.com/scipy/scipy/blob/9c0d08d7d11fc33311a96d2ac3ad73c8f6e3df00/scipy/ndimage/src/ni_interpolation.c#L412-L419 # noqa: E501
        # Also see this discussion:
        # https://github.com/dask/dask-image/issues/24#issuecomment-706165593 # noqa: E501
        for dim in range(n):

            if order % 2 == 0:
                rel_image_i[dim] += 0.5
                rel_image_f[dim] += 0.5

            rel_image_i[dim] = np.floor(rel_image_i[dim]) - order // 2
            rel_image_f[dim] = np.floor(rel_image_f[dim]) - order // 2 + order

            if order == 0:  # required for consistency with scipy.ndimage
                rel_image_i[dim] -= 1

        # clip image coordinates to image extent
        for dim, s in zip(range(n), image_shape):
            rel_image_i[dim] = np.clip(rel_image_i[dim], 0, s - 1)
            rel_image_f[dim] = np.clip(rel_image_f[dim], 0, s - 1)

        rel_image_slice = tuple([slice(int(rel_image_i[dim]),
                                       int(rel_image_f[dim]) + 2)
                                 for dim in range(n)])

        rel_image = image[rel_image_slice]

        """Block comment for future developers explaining how `offset` is
        transformed into `offset_prime` for each output chunk.
        Modify offset to point into cropped image.
        y = Mx + o
        Coordinate substitution:
        y' = y - y0(min_coord_px)
        x' = x - x0(chunk_offset)
        Then:
        y' = Mx' + o + Mx0 - y0
        M' = M
        o' = o + Mx0 - y0
        """

        offset_prime = offset + np.dot(matrix, out_chunk_offset) - rel_image_i

        output_layer[(output_name,) + block_ind] = (
                        affine_transform_method,
                        (da.core.concatenate3, rel_image.__dask_keys__()),
                        asarray_method(matrix),
                        offset_prime,
                        tuple(out_chunk_shape),  # output_shape
                        None,  # out
                        order,
                        mode,
                        cval,
                        False  # prefilter
        )

        rel_images.append(rel_image)

    graph = HighLevelGraph.from_collections(output_name, output_layer,
                                            dependencies=[image] + rel_images)

    meta = dispatch_asarray(image)([0]).astype(image.dtype)

    transformed = da.Array(graph,
                           output_name,
                           shape=tuple(output_shape),
                           # chunks=output_chunks,
                           chunks=normalized_chunks,
                           meta=meta)

    return transformed


# magnitude of the maximum filter pole for each order
# (obtained from scipy/ndimage/src/ni_splines.c)
_maximum_pole = {
    2: 0.171572875253809902396622551580603843,
    3: 0.267949192431122706472553658494127633,
    4: 0.361341225900220177092212841325675255,
    5: 0.430575347099973791851434783493520110,
}


def _get_default_depth(order, tol=1e-8):
    """Determine the approximate depth needed for a given tolerance.

    Here depth is chosen as the smallest integer such that ``|p| ** n < tol``
    where `|p|` is the magnitude of the largest pole in the IIR filter.
    """
    return math.ceil(np.log(tol) / np.log(_maximum_pole[order]))


def spline_filter(
        image,
        order=3,
        output=np.float64,
        mode='mirror',
        output_chunks=None,
        *,
        depth=None,
        **kwargs
):

    if not type(image) == da.core.Array:
        image = da.from_array(image)

    # use dispatching mechanism to determine backend
    spline_filter_method = dispatch_spline_filter(image)

    try:
        dtype = np.dtype(output)
    except TypeError:     # pragma: no cover
        raise TypeError(  # pragma: no cover
            "Could not coerce the provided output to a dtype. "
            "Passing array to output is not currently supported."
        )

    if depth is None:
        depth = _get_default_depth(order)

    if mode == 'wrap':
        raise NotImplementedError(
            "mode='wrap' is unsupported. It is recommended to use 'grid-wrap' "
            "instead."
        )

    # Note: depths of 12 and 24 give results matching SciPy to approximately
    #       single and double precision accuracy, respectively.
    boundary = "periodic" if mode == 'grid-wrap' else "none"
    depth, boundary = _get_depth_boundary(image.ndim, depth, boundary)

    # cannot pass a func kwarg named "output" to map_overlap
    spline_filter_method = functools.partial(spline_filter_method,
                                             output=dtype)

    result = image.map_overlap(
        spline_filter_method,
        depth=depth,
        boundary=boundary,
        dtype=dtype,
        meta=image._meta,
        # spline_filter kwargs
        order=order,
        mode=mode,
    )

    return result


def spline_filter1d(
        image,
        order=3,
        axis=-1,
        output=np.float64,
        mode='mirror',
        output_chunks=None,
        *,
        depth=None,
        **kwargs
):

    if not type(image) == da.core.Array:
        image = da.from_array(image)

    # use dispatching mechanism to determine backend
    spline_filter1d_method = dispatch_spline_filter1d(image)

    try:
        dtype = np.dtype(output)
    except TypeError:     # pragma: no cover
        raise TypeError(  # pragma: no cover
            "Could not coerce the provided output to a dtype. "
            "Passing array to output is not currently supported."
        )

    if depth is None:
        depth = _get_default_depth(order)

    # use depth 0 on all axes except the filtered axis
    if not np.isscalar(depth):
        raise ValueError("depth must be a scalar value")
    depths = [0] * image.ndim
    depths[axis] = depth

    if mode == 'wrap':
        raise NotImplementedError(
            "mode='wrap' is unsupported. It is recommended to use 'grid-wrap' "
            "instead."
        )

    # cannot pass a func kwarg named "output" to map_overlap
    spline_filter1d_method = functools.partial(spline_filter1d_method,
                                               output=dtype)

    result = image.map_overlap(
        spline_filter1d_method,
        depth=tuple(depths),
        boundary="periodic" if mode == 'grid-wrap' else "none",
        dtype=dtype,
        meta=image._meta,
        # spline_filter1d kwargs
        order=order,
        axis=axis,
        mode=mode,
    )

    return result



# -*- coding: utf-8 -*-
import glob
import numbers
import warnings

import dask.array as da
import numpy as np
import pims
from tifffile import natural_sorted


def imread(fname, nframes=1, *, arraytype="numpy"):
    """
    Read image data into a Dask Array.

    Provides a simple, fast mechanism to ingest image data into a
    Dask Array.

    Parameters
    ----------
    fname : str or pathlib.Path
        A glob like string that may match one or multiple filenames.
        Where multiple filenames match, they are sorted using
        natural (as opposed to alphabetical) sort.
    nframes : int, optional
        Number of the frames to include in each chunk (default: 1).
    arraytype : str, optional
        Array type for dask chunks. Available options: "numpy", "cupy".

    Returns
    -------
    array : dask.array.Array
        A Dask Array representing the contents of all image files.
    """

    sfname = str(fname)
    if not isinstance(nframes, numbers.Integral):
        raise ValueError("`nframes` must be an integer.")
    if (nframes != -1) and not (nframes > 0):
        raise ValueError("`nframes` must be greater than zero.")

    if arraytype == "numpy":
        arrayfunc = np.asanyarray
    elif arraytype == "cupy":   # pragma: no cover
        import cupy
        arrayfunc = cupy.asanyarray

    with pims.open(sfname) as imgs:
        shape = (len(imgs),) + imgs.frame_shape
        dtype = np.dtype(imgs.pixel_type)

    if nframes == -1:
        nframes = shape[0]

    if nframes > shape[0]:
        warnings.warn(
            "`nframes` larger than number of frames in file."
            " Will truncate to number of frames in file.",
            RuntimeWarning
        )
    elif shape[0] % nframes != 0:
        warnings.warn(
            "`nframes` does not nicely divide number of frames in file."
            " Last chunk will contain the remainder.",
            RuntimeWarning
        )

    # place source filenames into dask array after sorting
    filenames = natural_sorted(glob.glob(sfname))
    if len(filenames) > 1:
        ar = da.from_array(filenames, chunks=(nframes,))
        multiple_files = True
    else:
        ar = da.from_array(filenames * shape[0], chunks=(nframes,))
        multiple_files = False

    # read in data using encoded filenames
    a = ar.map_blocks(
        _map_read_frame,
        chunks=da.core.normalize_chunks(
            (nframes,) + shape[1:], shape),
        multiple_files=multiple_files,
        new_axis=list(range(1, len(shape))),
        arrayfunc=arrayfunc,
        meta=arrayfunc([]).astype(dtype),  # meta overwrites `dtype` argument
    )
    return a


def _map_read_frame(x, multiple_files, block_info=None, **kwargs):

    fn = x[0]  # get filename from input chunk

    if multiple_files:
        i, j = 0, 1
    else:
        i, j = block_info[None]['array-location'][0]

    return _read_frame(fn=fn, i=slice(i, j), **kwargs)


def _read_frame(fn, i, *, arrayfunc=np.asanyarray):
    with pims.open(fn) as imgs:
        return arrayfunc(imgs[i])



# -*- coding: utf-8 -*-
import scipy.ndimage

from ..dispatch._dispatch_ndmorph import (dispatch_binary_dilation,
                                          dispatch_binary_erosion)
from . import _ops, _utils

__all__ = [
    "binary_closing",
    "binary_dilation",
    "binary_erosion",
    "binary_opening",
]


@_utils._update_wrapper(scipy.ndimage.binary_closing)
def binary_closing(image,
                   structure=None,
                   iterations=1,
                   origin=0,
                   mask=None,
                   border_value=0,
                   brute_force=False):
    image = (image != 0)

    structure = _utils._get_structure(image, structure)
    iterations = _utils._get_iterations(iterations)
    origin = _utils._get_origin(structure.shape, origin)

    kwargs = dict(
        structure=structure,
        iterations=iterations,
        origin=origin,
        mask=mask,
        border_value=border_value,
        brute_force=brute_force
    )

    result = image
    result = binary_dilation(result, **kwargs)
    result = binary_erosion(result, **kwargs)

    return result


@_utils._update_wrapper(scipy.ndimage.binary_dilation)
def binary_dilation(image,
                    structure=None,
                    iterations=1,
                    mask=None,
                    border_value=0,
                    origin=0,
                    brute_force=False):
    border_value = _utils._get_border_value(border_value)

    result = _ops._binary_op(
        dispatch_binary_dilation(image),
        image,
        structure=structure,
        iterations=iterations,
        mask=mask,
        origin=origin,
        brute_force=brute_force,
        border_value=border_value
    )

    return result


@_utils._update_wrapper(scipy.ndimage.binary_erosion)
def binary_erosion(image,
                   structure=None,
                   iterations=1,
                   mask=None,
                   border_value=0,
                   origin=0,
                   brute_force=False):
    border_value = _utils._get_border_value(border_value)

    result = _ops._binary_op(
        dispatch_binary_erosion(image),
        image,
        structure=structure,
        iterations=iterations,
        mask=mask,
        origin=origin,
        brute_force=brute_force,
        border_value=border_value
    )

    return result


@_utils._update_wrapper(scipy.ndimage.binary_opening)
def binary_opening(image,
                   structure=None,
                   iterations=1,
                   origin=0,
                   mask=None,
                   border_value=0,
                   brute_force=False):
    image = (image != 0)

    structure = _utils._get_structure(image, structure)
    iterations = _utils._get_iterations(iterations)
    origin = _utils._get_origin(structure.shape, origin)

    kwargs = dict(
        structure=structure,
        iterations=iterations,
        origin=origin,
        mask=mask,
        border_value=border_value,
        brute_force=brute_force
    )

    result = image
    result = binary_erosion(result, **kwargs)
    result = binary_dilation(result, **kwargs)

    return result



# -*- coding: utf-8 -*-
import scipy.ndimage

from ..dispatch._dispatch_ndfilters import (dispatch_convolve,
                                            dispatch_correlate)
from ..dispatch._utils import check_arraytypes_compatible
from . import _utils

__all__ = [
    "convolve",
    "correlate",
]


@_utils._update_wrapper(scipy.ndimage.convolve)
def convolve(image, weights, mode="reflect", cval=0.0, origin=0):
    check_arraytypes_compatible(image, weights)

    origin = _utils._get_origin(weights.shape, origin)
    depth = _utils._get_depth(weights.shape, origin)
    depth, boundary = _utils._get_depth_boundary(image.ndim, depth, "none")

    if mode == "wrap":  # Fixes https://github.com/dask/dask-image/issues/242
        boundary = "periodic"
        mode = "constant"

    result = image.map_overlap(
        dispatch_convolve(image),
        depth=depth,
        boundary=boundary,
        dtype=image.dtype,
        meta=image._meta,
        weights=weights,
        mode=mode,
        cval=cval,
        origin=origin,
    )

    return result


@_utils._update_wrapper(scipy.ndimage.correlate)
def correlate(image, weights, mode="reflect", cval=0.0, origin=0):
    check_arraytypes_compatible(image, weights)

    origin = _utils._get_origin(weights.shape, origin)
    depth = _utils._get_depth(weights.shape, origin)
    depth, boundary = _utils._get_depth_boundary(image.ndim, depth, "none")

    if mode == "wrap":  # Fixes https://github.com/dask/dask-image/issues/242
        boundary = "periodic"
        mode = "constant"

    result = image.map_overlap(
        dispatch_correlate(image),
        depth=depth,
        boundary=boundary,
        dtype=image.dtype,
        meta=image._meta,
        weights=weights,
        mode=mode,
        cval=cval,
        origin=origin,
    )

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import numpy as np
import dask.array as da

import dask_image.ndmeasure._utils


def test__norm_input_labels_index_err():
    shape = (15, 16)
    chunks = (4, 5)
    ind = None

    a = np.random.random(shape)
    d = da.from_array(a, chunks=chunks)

    lbls = (a < 0.5).astype(np.int64)
    d_lbls = da.from_array(lbls, chunks=d.chunks)

    lbls = lbls[:-1]
    d_lbls = d_lbls[:-1]

    with pytest.raises(ValueError):
        dask_image.ndmeasure._utils._norm_input_labels_index(d, d_lbls, ind)


def test__norm_input_labels_index():
    shape = (15, 16)
    chunks = (4, 5)
    ind = None

    a = np.random.random(shape)
    d = da.from_array(a, chunks=chunks)

    lbls = (a < 0.5).astype(int)
    d_lbls = da.from_array(lbls, chunks=d.chunks)

    d_n, d_lbls_n, ind_n = dask_image.ndmeasure._utils._norm_input_labels_index(  # noqa: E501
        d, d_lbls, ind
    )

    assert isinstance(d_n, da.Array)
    assert isinstance(d_lbls_n, da.Array)
    assert isinstance(ind_n, da.Array)

    assert d_n.shape == d.shape
    assert d_lbls_n.shape == d_lbls.shape
    assert ind_n.shape == ()

    da.utils.assert_eq(d_n, d)
    da.utils.assert_eq(d_lbls_n, d_lbls)
    da.utils.assert_eq(ind_n, np.array(1, dtype=int))


@pytest.mark.parametrize(
    "shape, chunks, ind", [
        ((15, 16), (4, 5), 1),
        ((15, 16), (4, 5), [1]),
        ((15, 16), (4, 5), [[1, 2, 3, 4]]),
        ((15, 16), (4, 5), [[1, 2], [3, 4]]),
        ((15, 16), (4, 5), [[[1], [2], [3], [4]]]),
    ]
)
def test__norm_input_labels_index_warn(shape, chunks, ind):
    a = np.random.random(shape)
    d = da.from_array(a, chunks=chunks)

    lbls = np.zeros(a.shape, dtype=np.int64)
    lbls += (
        (a < 0.5).astype(lbls.dtype) +
        (a < 0.25).astype(lbls.dtype) +
        (a < 0.125).astype(lbls.dtype) +
        (a < 0.0625).astype(lbls.dtype)
    )
    d_lbls = da.from_array(lbls, chunks=d.chunks)

    ind = np.array(ind)
    d_ind = da.from_array(ind, chunks=1)

    with pytest.warns(None) as w:
        dask_image.ndmeasure._utils._norm_input_labels_index(
            d, d_lbls, d_ind
        )

    if ind.ndim > 1:
        assert len(w) == 1
        w.pop(FutureWarning)
    else:
        assert len(w) == 0


@pytest.mark.parametrize(
    "shape, chunks", [
        ((15,), (4,)),
        ((15, 16), (4, 5)),
        ((15, 1, 16), (4, 1, 5)),
        ((15, 12, 16), (4, 5, 6)),
    ]
)
def test___ravel_shape_indices(shape, chunks):
    a = np.arange(int(np.prod(shape)), dtype=np.int64).reshape(shape)
    d = dask_image.ndmeasure._utils._ravel_shape_indices(
        shape, dtype=np.int64, chunks=chunks
    )

    da.utils.assert_eq(d, a)



# -*- coding: utf-8 -*-

from ._dispatcher import get_type

__all__ = [
    "check_arraytypes_compatible",
]


def check_arraytypes_compatible(*args):
    """Check array types are compatible.

    For arrays to be compatible they must either have the same type,
    or a dask array where the chunks match the same array type.

    Examples of compatible arrays:
    * Two (or more) numpy arrays
    * A dask array with numpy chunks, and a numpy array

    Examples of incompatible arrays:
    * A numpy array and a cupy array
    """
    arraytypes = [get_type(arg) for arg in args]
    if len(set(arraytypes)) != 1:
        raise ValueError("Array types must be compatible.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numbers
from distutils.version import LooseVersion

import pytest
import numpy as np
import scipy as sp
import scipy.ndimage

import dask.array as da
import dask_image.ndfourier


@pytest.mark.parametrize(
    "err_type, s, n",
    [
        (NotImplementedError, 0.0, 0),
        (TypeError, 0.0 + 0.0j, 0),
        (TypeError, {}, 0),
        (RuntimeError, [0.0], 0),
        (RuntimeError, [[0.0], [0.0]], 0),
        (TypeError, [0.0, 0.0 + 0.0j], 0),
        (NotImplementedError, 0, 0),
    ]
)
@pytest.mark.parametrize(
    "funcname",
    [
        "fourier_shift",
        "fourier_gaussian",
        "fourier_uniform",
    ]
)
def test_fourier_filter_err(funcname, err_type, s, n):
    da_func = getattr(dask_image.ndfourier, funcname)

    a = np.arange(140.0).reshape(10, 14).astype(complex)
    d = da.from_array(a, chunks=(5, 7))

    with pytest.raises(err_type):
        da_func(d, s, n)


@pytest.mark.parametrize(
    "s",
    [
        0,
        (0, 0),
    ]
)
@pytest.mark.parametrize(
    "funcname",
    [
        "fourier_shift",
        "fourier_gaussian",
    ]
)
def test_fourier_filter_identity(funcname, s):
    da_func = getattr(dask_image.ndfourier, funcname)
    sp_func = getattr(scipy.ndimage, funcname)

    a = np.arange(140.0).reshape(10, 14).astype(complex)
    d = da.from_array(a, chunks=(5, 7))

    r_a = sp_func(a, s)
    r_d = da_func(d, s)

    assert d.chunks == r_d.chunks

    da.utils.assert_eq(d, r_d)
    da.utils.assert_eq(r_a, r_d)


@pytest.mark.parametrize(
    "dtype",
    [
        np.int64,
        np.float32,
        np.float64,
        np.complex64,
        np.complex128,
    ]
)
@pytest.mark.parametrize(
    "funcname, upcast_type",
    [
        ("fourier_shift", numbers.Real),
        ("fourier_gaussian", numbers.Integral),
        ("fourier_uniform", numbers.Integral),
    ]
)
def test_fourier_filter_type(funcname, upcast_type, dtype):
    if (
            LooseVersion(sp.__version__) >= "1.0.0" and
            dtype in [np.int64, np.float64] and
            funcname in ["fourier_gaussian", "fourier_uniform"]
       ):
        pytest.skip(
            "SciPy 1.0.0+ doesn't handle double precision values correctly."
        )

    dtype = np.dtype(dtype).type

    s = 1

    da_func = getattr(dask_image.ndfourier, funcname)
    sp_func = getattr(scipy.ndimage, funcname)

    a = np.arange(140.0).reshape(10, 14).astype(dtype)
    d = da.from_array(a, chunks=(5, 7))

    r_a = sp_func(a, s)
    r_d = da_func(d, s)

    assert d.chunks == r_d.chunks

    da.utils.assert_eq(r_a, r_d)

    if issubclass(dtype, upcast_type):
        assert r_d.real.dtype.type is np.float64
    else:
        assert r_d.dtype.type is dtype


@pytest.mark.parametrize(
    "shape, chunks",
    [
        ((10, 14), (10, 14)),
        ((10, 14), (5, 7)),
        ((10, 14), (6, 8)),
        ((10, 14), (4, 6)),
        ((16,), (3, 6, 2, 5)),
    ]
)
@pytest.mark.parametrize(
    "funcname",
    [
        "fourier_shift",
        "fourier_gaussian",
        "fourier_uniform",
    ]
)
def test_fourier_filter_chunks(funcname, shape, chunks):
    dtype = np.dtype(complex).type

    s = 1

    da_func = getattr(dask_image.ndfourier, funcname)
    sp_func = getattr(scipy.ndimage, funcname)

    a = np.arange(np.prod(shape)).reshape(shape).astype(dtype)
    d = da.from_array(a, chunks=chunks)

    r_a = sp_func(a, s)
    r_d = da_func(d, s)

    assert d.chunks == r_d.chunks

    da.utils.assert_eq(r_a, r_d)


@pytest.mark.parametrize(
    "s",
    [
        -1,
        (-1, -1),
        (-1, 2),
        (10, -9),
        (1, 0),
        (0, 2),
    ]
)
@pytest.mark.parametrize(
    "funcname",
    [
        "fourier_shift",
        "fourier_gaussian",
    ]
)
def test_fourier_filter_non_positive(funcname, s):
    da_func = getattr(dask_image.ndfourier, funcname)
    sp_func = getattr(scipy.ndimage, funcname)

    a = np.arange(140.0).reshape(10, 14).astype(complex)
    d = da.from_array(a, chunks=(5, 7))

    r_a = sp_func(a, s)
    r_d = da_func(d, s)

    assert d.chunks == r_d.chunks

    da.utils.assert_eq(r_a, r_d)


@pytest.mark.parametrize(
    "s",
    [
        1,
        0.5,
        (1, 1),
        (0.8, 1.5),
        np.ones((2,)),
        da.ones((2,), chunks=(2,)),
        da.ones((2,), chunks=(1,)),
    ]
)
@pytest.mark.parametrize(
    "funcname",
    [
        "fourier_shift",
        "fourier_gaussian",
        "fourier_uniform",
    ]
)


@pytest.mark.parametrize(
    "real_fft, axis",
    [
        (True, -1),
        (True, 0),
        (False, -1),
    ]
)
def test_fourier_filter(funcname, s, real_fft, axis):
    da_func = getattr(dask_image.ndfourier, funcname)
    sp_func = getattr(scipy.ndimage, funcname)

    shape = (10, 14)
    n = 2 * shape[axis] - 1 if real_fft else -1
    dtype = np.float64 if real_fft else np.complex128

    a = np.arange(140.0).reshape(shape).astype(dtype)
    d = da.from_array(a, chunks=(5, 7))

    r_a = sp_func(a, s, n=n, axis=axis)
    r_d = da_func(d, s, n=n, axis=axis)

    assert d.chunks == r_d.chunks

    da.utils.assert_eq(r_a, r_d)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import numpy as np
import scipy.ndimage

import dask.array as da

import dask_image.ndfilters


@pytest.mark.parametrize(
    "err_type, size, origin",
    [
        (TypeError, 3.0, 0),
        (TypeError, 3, 0.0),
        (RuntimeError, [3], 0),
        (RuntimeError, 3, [0]),
        (RuntimeError, [[3]], 0),
        (RuntimeError, 3, [[0]]),
    ]
)
def test_uniform_filter_params(err_type, size, origin):
    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    with pytest.raises(err_type):
        dask_image.ndfilters.uniform_filter(d, size, origin=origin)


def test_uniform_shape_type():
    size = 1
    origin = 0

    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    assert all([(type(s) is int) for s in d.shape])

    d2 = dask_image.ndfilters.uniform_filter(d, size, origin=origin)

    assert all([(type(s) is int) for s in d2.shape])


def test_uniform_comprehensions():
    da_func = lambda arr: dask_image.ndfilters.uniform_filter(arr, 1, origin=0)  # noqa: E731, E501

    np.random.seed(0)

    a = np.random.random((3, 12, 14))
    d = da.from_array(a, chunks=(3, 6, 7))

    l2s = [da_func(d[i]) for i in range(len(d))]
    l2c = [da_func(d[i])[None] for i in range(len(d))]

    da.utils.assert_eq(np.stack(l2s), da.stack(l2s))
    da.utils.assert_eq(np.concatenate(l2c), da.concatenate(l2c))


@pytest.mark.parametrize(
    "size, origin",
    [
        (1, 0),
    ]
)
def test_uniform_identity(size, origin):
    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    da.utils.assert_eq(
        d, dask_image.ndfilters.uniform_filter(d, size, origin=origin)
    )

    da.utils.assert_eq(
        scipy.ndimage.uniform_filter(a, size, origin=origin),
        dask_image.ndfilters.uniform_filter(d, size, origin=origin)
    )


@pytest.mark.parametrize(
    "size, origin",
    [
        (2, 0),
        (3, 0),
        (3, 1),
        (3, (1, 0)),
        ((1, 2), 0),
        ((3, 2), (1, 0)),
    ]
)
def test_uniform_compare(size, origin):
    s = (100, 110)
    a = np.arange(float(np.prod(s))).reshape(s)
    d = da.from_array(a, chunks=(50, 55))

    da.utils.assert_eq(
        scipy.ndimage.uniform_filter(a, size, origin=origin),
        dask_image.ndfilters.uniform_filter(d, size, origin=origin)
    )



# -*- coding: utf-8 -*-


import numbers

import scipy.ndimage

from ..dispatch._dispatch_ndfilters import dispatch_prewitt, dispatch_sobel
from . import _utils

__all__ = [
    "prewitt",
    "sobel",
]


def _validate_axis(ndim, axis):
    if not isinstance(axis, numbers.Integral):
        raise ValueError("The axis must be of integral type.")
    if axis < -ndim or axis >= ndim:
        raise ValueError("The axis is out of range.")


@_utils._update_wrapper(scipy.ndimage.prewitt)
def prewitt(image, axis=-1, mode='reflect', cval=0.0):
    _validate_axis(image.ndim, axis)

    result = image.map_overlap(
        dispatch_prewitt(image),
        depth=(image.ndim * (1,)),
        boundary="none",
        dtype=image.dtype,
        meta=image._meta,
        axis=axis,
        mode=mode,
        cval=cval
    )

    return result


@_utils._update_wrapper(scipy.ndimage.sobel)
def sobel(image, axis=-1, mode='reflect', cval=0.0):
    _validate_axis(image.ndim, axis)

    result = image.map_overlap(
        dispatch_sobel(image),
        depth=(image.ndim * (1,)),
        boundary="none",
        dtype=image.dtype,
        meta=image._meta,
        axis=axis,
        mode=mode,
        cval=cval
    )

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from packaging import version

import dask
import dask.array as da
import numpy as np
import pytest
import scipy
import scipy.ndimage

import dask_image.ndinterp

# mode lists for the case with prefilter = False
_supported_modes = ['constant', 'nearest']
_unsupported_modes = ['wrap', 'reflect', 'mirror']

# mode lists for the case with prefilter = True
_supported_prefilter_modes = ['constant']
_unsupported_prefilter_modes = _unsupported_modes + ['nearest']

have_scipy16 = version.parse(scipy.__version__) >= version.parse('1.6.0')

# additional modes are present in SciPy >= 1.6.0
if have_scipy16:
    _supported_modes += ['grid-constant']
    _unsupported_modes += ['grid-mirror', 'grid-wrap']
    _unsupported_prefilter_modes += ['grid-constant', 'grid-mirror',
                                     'grid-wrap']


def validate_affine_transform(n=2,
                              matrix=None,
                              offset=None,
                              input_output_shape_per_dim=(16, 16),
                              interp_order=1,
                              interp_mode='constant',
                              input_output_chunksize_per_dim=(6, 6),
                              random_seed=0,
                              use_cupy=False,
                              prefilter=False
                              ):
    """
    Compare the outputs of `scipy.ndimage.affine_transformation`
    and `dask_image.ndinterp.affine_transformation`.

    Notes
    -----
        Currently, prefilter is disabled and therefore the output
        of `dask_image.ndinterp.affine_transformation` is compared
        to `prefilter=False`.
    """

    if (interp_order > 1 and interp_mode == 'nearest' and not have_scipy16):
        # not clear on the underlying cause, but this fails on older SciPy
        pytest.skip("requires SciPy >= 1.6.0")

    # define test image
    a = input_output_shape_per_dim[0]
    np.random.seed(random_seed)
    image = np.random.random([a] * n)

    # transform into dask array
    chunksize = [input_output_chunksize_per_dim[0]] * n
    image_da = da.from_array(image, chunks=chunksize)
    if use_cupy:
        import cupy as cp
        image_da = image_da.map_blocks(cp.asarray)

    if (
        prefilter
        and interp_mode in _supported_prefilter_modes
        and interp_order > 1
        and version.parse(dask.__version__) < version.parse("2020.1.0")
    ):
        # older dask will fail if any chunks have size smaller than depth
        depth = dask_image.ndinterp._get_default_depth(interp_order)
        in_size = input_output_shape_per_dim[0]
        in_chunksize = input_output_chunksize_per_dim[0]
        rem = in_size % in_chunksize
        if in_size < depth or (rem != 0 and rem < depth):
            pytest.skip("older dask doesn't automatically rechunk")

    # define (random) transformation
    if matrix is None:
        # make sure to substantially deviate from unity matrix
        matrix = np.eye(n) + (np.random.random((n, n)) - 0.5) * 5.
    if offset is None:
        offset = (np.random.random(n) - 0.5) / 5. * np.array(image.shape)

    # define resampling options
    output_shape = [input_output_shape_per_dim[1]] * n
    output_chunks = [input_output_chunksize_per_dim[1]] * n

    # transform with scipy
    image_t_scipy = scipy.ndimage.affine_transform(
        image, matrix, offset,
        output_shape=output_shape,
        order=interp_order,
        mode=interp_mode,
        prefilter=prefilter)

    # transform with dask-image
    image_t_dask = dask_image.ndinterp.affine_transform(
        image_da, matrix, offset,
        output_shape=output_shape,
        output_chunks=output_chunks,
        order=interp_order,
        mode=interp_mode,
        prefilter=prefilter)
    image_t_dask_computed = image_t_dask.compute()

    assert np.allclose(image_t_scipy, image_t_dask_computed)


@pytest.mark.parametrize("n",
                         [1, 2, 3])
@pytest.mark.parametrize("input_output_shape_per_dim",
                         [(25, 25)])
@pytest.mark.parametrize("interp_order",
                         range(6))
@pytest.mark.parametrize("input_output_chunksize_per_dim",
                         [(16, 16), (16, 7), (7, 16)])
@pytest.mark.parametrize("random_seed",
                         [0, 2])
def test_affine_transform_general(n,
                                  input_output_shape_per_dim,
                                  interp_order,
                                  input_output_chunksize_per_dim,
                                  random_seed):

    kwargs = dict()
    kwargs['n'] = n
    kwargs['input_output_shape_per_dim'] = input_output_shape_per_dim
    kwargs['interp_order'] = interp_order
    kwargs['input_output_chunksize_per_dim'] = input_output_chunksize_per_dim
    kwargs['random_seed'] = random_seed

    validate_affine_transform(**kwargs)


@pytest.mark.cupy
@pytest.mark.parametrize("n",
                         [1, 2, 3])
@pytest.mark.parametrize("input_output_shape_per_dim",
                         [(25, 25), (25, 10)])
@pytest.mark.parametrize("interp_order",
                         [0, 1])
@pytest.mark.parametrize("input_output_chunksize_per_dim",
                         [(16, 16), (16, 7)])
@pytest.mark.parametrize("random_seed",
                         [0])
def test_affine_transform_cupy(n,
                               input_output_shape_per_dim,
                               interp_order,
                               input_output_chunksize_per_dim,
                               random_seed):
    pytest.importorskip("cupy", minversion="5.0.0")

    kwargs = dict()
    kwargs['n'] = n
    kwargs['input_output_shape_per_dim'] = input_output_shape_per_dim
    kwargs['interp_order'] = interp_order
    kwargs['input_output_chunksize_per_dim'] = input_output_chunksize_per_dim
    kwargs['random_seed'] = random_seed
    kwargs['use_cupy'] = True

    validate_affine_transform(**kwargs)


@pytest.mark.parametrize("n",
                         [1, 2, 3])
@pytest.mark.parametrize("interp_mode",
                         _supported_modes)
@pytest.mark.parametrize("interp_order",
                         [0, 3])
@pytest.mark.parametrize("input_output_shape_per_dim",
                         [(20, 30)])
@pytest.mark.parametrize("input_output_chunksize_per_dim",
                         [(15, 10)])
def test_affine_transform_modes(n,
                                interp_mode,
                                interp_order,
                                input_output_shape_per_dim,
                                input_output_chunksize_per_dim,
                                ):

    kwargs = dict()
    kwargs['n'] = n
    kwargs['interp_mode'] = interp_mode
    kwargs['input_output_shape_per_dim'] = input_output_shape_per_dim
    kwargs['input_output_chunksize_per_dim'] = input_output_chunksize_per_dim
    kwargs['interp_order'] = interp_order
    kwargs['prefilter'] = False

    validate_affine_transform(**kwargs)


@pytest.mark.parametrize("interp_mode",
                         _unsupported_modes)
def test_affine_transform_unsupported_modes(interp_mode):

    with pytest.raises(NotImplementedError):
        validate_affine_transform(interp_mode=interp_mode)


@pytest.mark.parametrize("n", [1, 2, 3])
@pytest.mark.parametrize("interp_order", range(6))
@pytest.mark.parametrize("interp_mode", _supported_prefilter_modes)
def test_affine_transform_prefilter_modes(n, interp_order, interp_mode):

    validate_affine_transform(
        n=n,
        input_output_shape_per_dim=(32, 32),
        input_output_chunksize_per_dim=(24, 24),
        interp_order=interp_order,
        interp_mode=interp_mode,
        prefilter=True,
    )


@pytest.mark.parametrize("n", [1, 2, 3])
@pytest.mark.parametrize("interp_order", range(2, 6))
@pytest.mark.parametrize("interp_mode", _unsupported_prefilter_modes)
def test_affine_transform_prefilter_not_implemented(
    n, interp_order, interp_mode
):

    with pytest.raises(NotImplementedError):
        validate_affine_transform(
            n=n,
            interp_order=interp_order,
            interp_mode=interp_mode,
            prefilter=True,
        )


def test_affine_transform_numpy_input():

    image = np.ones((3, 3))
    image_t = dask_image.ndinterp.affine_transform(image, np.eye(2), [0, 0])

    assert image_t.shape == image.shape
    assert (image == image_t).min()


def test_affine_transform_minimal_input():

    image = np.ones((3, 3))
    image_t = dask_image.ndinterp.affine_transform(np.ones((3, 3)), np.eye(2))

    assert image_t.shape == image.shape


def test_affine_transform_type_consistency():

    image = da.ones((3, 3))
    image_t = dask_image.ndinterp.affine_transform(image, np.eye(2), [0, 0])

    assert isinstance(image, type(image_t))
    assert isinstance(image[0, 0].compute(), type(image_t[0, 0].compute()))


@pytest.mark.cupy
def test_affine_transform_type_consistency_gpu():

    cupy = pytest.importorskip("cupy", minversion="5.0.0")

    image = da.ones((3, 3))
    image_t = dask_image.ndinterp.affine_transform(image, np.eye(2), [0, 0])

    image.map_blocks(cupy.asarray)

    assert isinstance(image, type(image_t))
    assert isinstance(image[0, 0].compute(), type(image_t[0, 0].compute()))


def test_affine_transform_no_output_shape_or_chunks_specified():

    image = da.ones((3, 3))
    image_t = dask_image.ndinterp.affine_transform(image, np.eye(2), [0, 0])

    assert image_t.shape == image.shape
    assert image_t.chunks == tuple([(s,) for s in image.shape])


def test_affine_transform_prefilter_warning():

    with pytest.warns(UserWarning):
        dask_image.ndinterp.affine_transform(da.ones(20), [1], [0],
                                             order=3, prefilter=True)


@pytest.mark.timeout(15)
def test_affine_transform_large_input_small_output_cpu():
    """
    Make sure input array does not need to be computed entirely
    """

    # fully computed, this array would occupy 8TB
    image = da.random.random([10000] * 3, chunks=(200, 200, 200))
    image_t = dask_image.ndinterp.affine_transform(image, np.eye(3), [0, 0, 0],
                                                   output_chunks=[1, 1, 1],
                                                   output_shape=[1, 1, 1])

    # if more than the needed chunks should be computed,
    # this would take long and eventually raise a MemoryError
    image_t[0, 0, 0].compute()


@pytest.mark.cupy
@pytest.mark.timeout(15)
def test_affine_transform_large_input_small_output_gpu():
    """
    Make sure input array does not need to be computed entirely
    """
    cupy = pytest.importorskip("cupy", minversion="5.0.0")

    # this array would occupy more than 24GB on a GPU
    image = da.random.random([2000] * 3, chunks=(50, 50, 50))
    image.map_blocks(cupy.asarray)

    image_t = dask_image.ndinterp.affine_transform(image, np.eye(3), [0, 0, 0],
                                                   output_chunks=[1, 1, 1],
                                                   output_shape=[1, 1, 1])
    # if more than the needed chunks should be computed,
    # this would take long and eventually raise a MemoryError
    image_t[0, 0, 0].compute()


@pytest.mark.filterwarnings("ignore:The behavior of affine_transform "
                            "with a 1-D array supplied for the matrix "
                            "parameter has changed")
@pytest.mark.parametrize("n",
                         [1, 2, 3, 4])
def test_affine_transform_parameter_formats(n):

    # define reference parameters
    scale_factors = np.ones(n, dtype=float) * 2.
    matrix_n = np.diag(scale_factors)
    offset = -np.ones(n)

    # convert into different formats
    matrix_only_scaling = scale_factors
    matrix_pre_homogeneous = np.hstack((matrix_n, offset[:, None]))
    matrix_homogeneous = np.vstack((matrix_pre_homogeneous,
                                   [0] * n + [1]))

    np.random.seed(0)
    image = da.random.random([5] * n)

    # reference run
    image_t_0 = dask_image.ndinterp.affine_transform(image,
                                                     matrix_n,
                                                     offset).compute()

    # assert that the different parameter formats
    # lead to the same output
    image_t_scale = dask_image.ndinterp.affine_transform(image,
                                                         matrix_only_scaling,
                                                         offset).compute()
    assert (np.allclose(image_t_0, image_t_scale))

    for matrix in [matrix_pre_homogeneous, matrix_homogeneous]:

        image_t = dask_image.ndinterp.affine_transform(image,
                                                       matrix,
                                                       offset + 10.,  # ignored
                                                       ).compute()

        assert(np.allclose(image_t_0, image_t))

    # catch matrices that are not homogeneous transformation matrices
    with pytest.raises(ValueError):
        matrix_not_homogeneous = np.vstack((matrix_pre_homogeneous,
                                           [-1] * n + [1]))
        dask_image.ndinterp.affine_transform(image,
                                             matrix_not_homogeneous,
                                             offset)



# -*- coding: utf-8 -*-

import numpy as np
import scipy.ndimage

from ._dispatcher import Dispatcher

__all__ = [
    "dispatch_convolve",
    "dispatch_correlate",
    "dispatch_laplace",
    "dispatch_prewitt",
    "dispatch_sobel",
    "dispatch_gaussian_filter",
    "dispatch_gaussian_gradient_magnitude",
    "dispatch_gaussian_laplace",
    "dispatch_generic_filter",
    "dispatch_minimum_filter",
    "dispatch_median_filter",
    "dispatch_maximum_filter",
    "dispatch_rank_filter",
    "dispatch_percentile_filter",
    "dispatch_uniform_filter",
    "dispatch_threshold_local_mean",
]


dispatch_convolve = Dispatcher(name="dispatch_convolve")
dispatch_correlate = Dispatcher(name="dispatch_correlate")
dispatch_laplace = Dispatcher(name="dispatch_laplace")
dispatch_prewitt = Dispatcher(name="dispatch_prewitt")
dispatch_sobel = Dispatcher(name="dispatch_sobel")
dispatch_gaussian_filter = Dispatcher(name="dispatch_gaussian_filter")
dispatch_gaussian_gradient_magnitude = Dispatcher(name="dispatch_gaussian_gradient_magnitude")  # noqa: E501
dispatch_gaussian_laplace = Dispatcher(name="dispatch_gaussian_laplace")
dispatch_generic_filter = Dispatcher(name="dispatch_generic_filter")
dispatch_minimum_filter = Dispatcher(name="dispatch_minimum_filter")
dispatch_median_filter = Dispatcher(name="dispatch_median_filter")
dispatch_maximum_filter = Dispatcher(name="dispatch_maximum_filter")
dispatch_rank_filter = Dispatcher(name="dispatch_rank_filter")
dispatch_percentile_filter = Dispatcher(name="dispatch_percentile_filter")
dispatch_uniform_filter = Dispatcher(name="dispatch_uniform_filter")
dispatch_threshold_local_mean = Dispatcher(name="dispatch_threshold_local_mean")  # noqa: E501


# ================== convolve ==================
@dispatch_convolve.register(np.ndarray)
def numpy_convolve(*args, **kwargs):
    return scipy.ndimage.convolve


@dispatch_convolve.register_lazy("cupy")
def register_cupy_convolve():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_convolve.register(cupy.ndarray)
    def cupy_convolve(*args, **kwargs):
        return cupyx.scipy.ndimage.convolve


# ================== correlate ==================
@dispatch_correlate.register(np.ndarray)
def numpy_correlate(*args, **kwargs):
    return scipy.ndimage.correlate


@dispatch_correlate.register_lazy("cupy")
def register_cupy_correlate():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_correlate.register(cupy.ndarray)
    def cupy_correlate(*args, **kwargs):
        return cupyx.scipy.ndimage.correlate


# ================== laplace ==================
@dispatch_laplace.register(np.ndarray)
def numpy_laplace(*args, **kwargs):
    return scipy.ndimage.laplace


@dispatch_laplace.register_lazy("cupy")
def register_cupy_laplace():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_laplace.register(cupy.ndarray)
    def cupy_laplace(*args, **kwargs):
        return cupyx.scipy.ndimage.laplace


# ================== prewitt ==================
@dispatch_prewitt.register(np.ndarray)
def numpy_prewitt(*args, **kwargs):
    return scipy.ndimage.prewitt


@dispatch_prewitt.register_lazy("cupy")
def register_cupy_prewitt():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_prewitt.register(cupy.ndarray)
    def cupy_prewitt(*args, **kwargs):
        return cupyx.scipy.ndimage.prewitt


# ================== sobel ==================
@dispatch_sobel.register(np.ndarray)
def numpy_sobel(*args, **kwargs):
    return scipy.ndimage.sobel


@dispatch_sobel.register_lazy("cupy")
def register_cupy_sobel():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_sobel.register(cupy.ndarray)
    def cupy_sobel(*args, **kwargs):
        return cupyx.scipy.ndimage.sobel


# ================== gaussian_filter ==================
@dispatch_gaussian_filter.register(np.ndarray)
def numpy_gaussian_filter(*args, **kwargs):
    return scipy.ndimage.gaussian_filter


@dispatch_gaussian_filter.register_lazy("cupy")
def register_cupy_gaussian_filter():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_gaussian_filter.register(cupy.ndarray)
    def cupy_gaussian_filter(*args, **kwargs):
        return cupyx.scipy.ndimage.gaussian_filter


# ================== gaussian_gradient_magnitude ==================
@dispatch_gaussian_gradient_magnitude.register(np.ndarray)
def numpy_gaussian_gradient_magnitude(*args, **kwargs):
    return scipy.ndimage.gaussian_gradient_magnitude


@dispatch_gaussian_gradient_magnitude.register_lazy("cupy")
def register_cupy_gaussian_gradient_magnitude():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_gaussian_gradient_magnitude.register(cupy.ndarray)
    def cupy_gaussian_gradient_magnitude(*args, **kwargs):
        return cupyx.scipy.ndimage.gaussian_gradient_magnitude


# ================== gaussian_laplace ==================
@dispatch_gaussian_laplace.register(np.ndarray)
def numpy_gaussian_laplace(*args, **kwargs):
    return scipy.ndimage.gaussian_laplace


@dispatch_gaussian_laplace.register_lazy("cupy")
def register_cupy_gaussian_laplace():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_gaussian_laplace.register(cupy.ndarray)
    def cupy_gaussian_laplace(*args, **kwargs):
        return cupyx.scipy.ndimage.gaussian_laplace


# ================== generic_filter ==================
@dispatch_generic_filter.register(np.ndarray)
def numpy_generic_filter(*args, **kwargs):
    return scipy.ndimage.generic_filter


@dispatch_generic_filter.register_lazy("cupy")
def register_cupy_generic_filter():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_generic_filter.register(cupy.ndarray)
    def cupy_generic_filter(*args, **kwargs):
        return cupyx.scipy.ndimage.generic_filter


# ================== minimum_filter ==================
@dispatch_minimum_filter.register(np.ndarray)
def numpy_minimum_filter(*args, **kwargs):
    return scipy.ndimage.minimum_filter


@dispatch_minimum_filter.register_lazy("cupy")
def register_cupy_minimum_filter():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_minimum_filter.register(cupy.ndarray)
    def cupy_minimum_filter(*args, **kwargs):
        return cupyx.scipy.ndimage.minimum_filter


# ================== median_filter ==================
@dispatch_median_filter.register(np.ndarray)
def numpy_median_filter(*args, **kwargs):
    return scipy.ndimage.median_filter


@dispatch_median_filter.register_lazy("cupy")
def register_cupy_median_filter():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_median_filter.register(cupy.ndarray)
    def cupy_median_filter(*args, **kwargs):
        return cupyx.scipy.ndimage.median_filter


# ================== maximum_filter ==================
@dispatch_maximum_filter.register(np.ndarray)
def numpy_maximum_filter(*args, **kwargs):
    return scipy.ndimage.maximum_filter


@dispatch_maximum_filter.register_lazy("cupy")
def register_cupy_maximum_filter():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_maximum_filter.register(cupy.ndarray)
    def cupy_maximum_filter(*args, **kwargs):
        return cupyx.scipy.ndimage.maximum_filter


# ================== rank_filter ==================
@dispatch_rank_filter.register(np.ndarray)
def numpy_rank_filter(*args, **kwargs):
    return scipy.ndimage.rank_filter


@dispatch_rank_filter.register_lazy("cupy")
def register_cupy_rank_filter():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_rank_filter.register(cupy.ndarray)
    def cupy_rank_filter(*args, **kwargs):
        return cupyx.scipy.ndimage.rank_filter


# ================== percentile_filter ==================
@dispatch_percentile_filter.register(np.ndarray)
def numpy_percentile_filter(*args, **kwargs):
    return scipy.ndimage.percentile_filter


@dispatch_percentile_filter.register_lazy("cupy")
def register_cupy_percentile_filter():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_percentile_filter.register(cupy.ndarray)
    def cupy_percentile_filter(*args, **kwargs):
        return cupyx.scipy.ndimage.percentile_filter


# ================== uniform_filter ==================
@dispatch_uniform_filter.register(np.ndarray)
def numpy_uniform_filter(*args, **kwargs):
    return scipy.ndimage.uniform_filter


@dispatch_uniform_filter.register_lazy("cupy")
def register_cupy_uniform_filter():
    import cupy
    import cupyx.scipy.ndimage

    @dispatch_uniform_filter.register(cupy.ndarray)
    def cupy_uniform_filter(*args, **kwargs):
        return cupyx.scipy.ndimage.uniform_filter


# ================== threshold_local_mean ==================
@dispatch_threshold_local_mean.register(np.ndarray)
def numpy_threshold_local_mean(*args, **kwargs):
    return np.mean


@dispatch_threshold_local_mean.register_lazy("cupy")
def register_cupy_threshold_local_mean():
    import cupy

    @dispatch_threshold_local_mean.register(cupy.ndarray)
    def cupy_threshold_local_mean(*args, **kwargs):
        # Code snippet taken from https://github.com/cupy/cupy/issues/3909
        my_mean = cupy.ReductionKernel(
            'T x',  # input params
            'T y',  # output params
            'x',  # map
            'a + b',  # reduce
            'y = a / _in_ind.size()',  # An undocumented variable and a hack
            '0',  # identity value
            'mean'  # kernel name
        )
        return my_mean



# -*- coding: utf-8 -*-

from dask.utils import Dispatch


def get_type(array):
    """Return type of arrays contained within the dask array chunks."""
    try:
        datatype = type(array._meta)  # Check chunk type backing dask array
    except AttributeError:
        datatype = type(array)  # For all non-dask arrays
    return datatype


class Dispatcher(Dispatch):
    """Simple single dispatch for different dask array types."""

    def __call__(self, arg, *args, **kwargs):
        """
        Call the corresponding method based on type of dask array.
        """
        datatype = get_type(arg)
        meth = self.dispatch(datatype)
        return meth(arg, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import scipy.ndimage

import dask.array as da

import dask_image.ndfilters


def test_laplace_comprehensions():
    np.random.seed(0)

    a = np.random.random((3, 12, 14))
    d = da.from_array(a, chunks=(3, 6, 7))

    l2s = [dask_image.ndfilters.laplace(d[i]) for i in range(len(d))]
    l2c = [dask_image.ndfilters.laplace(d[i])[None] for i in range(len(d))]

    da.utils.assert_eq(np.stack(l2s), da.stack(l2s))
    da.utils.assert_eq(np.concatenate(l2c), da.concatenate(l2c))


def test_laplace_compare():
    s = (10, 11, 12)
    a = np.arange(float(np.prod(s))).reshape(s)
    d = da.from_array(a, chunks=(5, 5, 6))

    da.utils.assert_eq(
        scipy.ndimage.laplace(a), dask_image.ndfilters.laplace(d)
    )



# -*- coding: utf-8 -*-

__all__ = [
    "convolve",
    "correlate",
    "laplace",
    "prewitt",
    "sobel",
    "gaussian",
    "gaussian_filter",
    "gaussian_gradient_magnitude",
    "gaussian_laplace",
    "generic_filter",
    "minimum_filter",
    "median_filter",
    "maximum_filter",
    "rank_filter",
    "percentile_filter",
    "uniform_filter",
    "threshold_local",
]

from ._conv import convolve, correlate
from ._diff import laplace
from ._edge import prewitt, sobel
from ._gaussian import (gaussian, gaussian_filter,
                        gaussian_gradient_magnitude, gaussian_laplace)
from ._generic import generic_filter
from ._order import (maximum_filter, median_filter, minimum_filter,
                     percentile_filter, rank_filter)
from ._smooth import uniform_filter
from ._threshold import threshold_local

convolve.__module__ = __name__
correlate.__module__ = __name__

laplace.__module__ = __name__

prewitt.__module__ = __name__
sobel.__module__ = __name__

gaussian.__module__ = __name__
gaussian_filter.__module__ = __name__
gaussian_gradient_magnitude.__module__ = __name__
gaussian_laplace.__module__ = __name__

generic_filter.__module__ = __name__

minimum_filter.__module__ = __name__
median_filter.__module__ = __name__
maximum_filter.__module__ = __name__
rank_filter.__module__ = __name__
percentile_filter.__module__ = __name__

uniform_filter.__module__ = __name__

threshold_local.__module__ = __name__



import dask.array as da
import numpy as np

from ..dispatch._dispatch_ndfilters import dispatch_threshold_local_mean
from . import _gaussian, _generic, _order

__all__ = [
    "threshold_local",
]


def threshold_local(image, block_size, method='gaussian', offset=0,
                    mode='reflect', param=None, cval=0):
    """Compute a threshold mask image based on local pixel neighborhood.

    Also known as adaptive or dynamic thresholding[1]_. The threshold value is
    the weighted mean for the local neighborhood of a pixel subtracted by a
    constant. Alternatively the threshold can be determined dynamically by a
    given function, using the 'generic' method.

    Parameters
    ----------
    image : (N, M) dask ndarray
        Input image.
    block_size : int or list/tuple/array
        Size of pixel neighborhood which is used to calculate the
        threshold value.
        (1) A single value for use in all dimensions or
        (2) A tuple, list, or array with length equal to image.ndim
    method : {'generic', 'gaussian', 'mean', 'median'}, optional
        Method used to determine adaptive threshold for local neighbourhood in
        weighted mean image.

        * 'generic': use custom function (see `param` parameter)
        * 'gaussian': apply gaussian filter (see `param` parameter for custom\
                      sigma value)
        * 'mean': apply arithmetic mean filter
        * 'median': apply median rank filter

        By default the 'gaussian' method is used.
    offset : float, optional
        Constant subtracted from weighted mean of neighborhood to calculate
        the local threshold value. Default offset is 0.
    mode : {'reflect', 'constant', 'nearest', 'mirror', 'wrap'}, optional
        The mode parameter determines how the array borders are handled, where
        cval is the value when mode is equal to 'constant'.
        Default is 'reflect'.
    param : {int, function}, optional
        Either specify sigma for 'gaussian' method or function object for
        'generic' method. This functions takes the flat array of local
        neighbourhood as a single argument and returns the calculated
        threshold for the centre pixel.
    cval : float, optional
        Value to fill past edges of input if mode is 'constant'.

    Returns
    -------
    threshold : (N, M) dask ndarray
        Threshold image. All pixels in the input image higher than the
        corresponding pixel in the threshold image are considered foreground.

    References
    ----------
    .. [1] https://docs.opencv.org/modules/imgproc/doc/miscellaneous_transformations.html?highlight=threshold

    Examples
    --------
    >>> import dask.array as da
    >>> image = da.random.random((1000, 1000), chunks=(100, 100))
    >>> result = threshold_local(image, 15, 'gaussian')
    """  # noqa

    image = image.astype(np.float64)

    if method == 'generic':
        if not callable(param):
            raise ValueError("Must include a valid function to use as the "
                             "'param' keyword argument.")
        thresh_image = _generic.generic_filter(image, param, block_size,
                                               mode=mode, cval=cval)
    elif method == 'gaussian':
        if param is None:
            sigma = (np.array(block_size).astype(float) - 1) / 6.0
        else:
            sigma = param
        thresh_image = _gaussian.gaussian_filter(image, sigma, mode=mode,
                                                 cval=cval)
    elif method == 'mean':
        thresh_image = _generic.generic_filter(
            image, dispatch_threshold_local_mean(image), block_size, mode=mode,
            cval=cval)
    elif method == 'median':
        thresh_image = _order.median_filter(image, block_size, mode=mode,
                                            cval=cval)
    else:
        raise ValueError("Invalid method specified. Please use `generic`, "
                         "`gaussian`, `mean`, or `median`.")
    return thresh_image - offset



# -*- coding: utf-8 -*-
import numbers

import dask.array as da

from . import _utils

__all__ = [
    "fourier_gaussian",
    "fourier_shift",
    "fourier_uniform",
]


def fourier_gaussian(image, sigma, n=-1, axis=-1):
    """
    Multi-dimensional Gaussian fourier filter.

    The array is multiplied with the fourier transform of a Gaussian
    kernel.

    Parameters
    ----------
    image : array_like
        The input image.
    sigma : float or sequence
        The sigma of the Gaussian kernel. If a float, `sigma` is the same for
        all axes. If a sequence, `sigma` has to contain one value for each
        axis.
    n : int, optional
        If `n` is negative (default), then the image is assumed to be the
        result of a complex fft.
        If `n` is larger than or equal to zero, the image is assumed to be the
        result of a real fft, and `n` gives the length of the array before
        transformation along the real transform direction.
    axis : int, optional
        The axis of the real transform.

    Returns
    -------
    fourier_gaussian : Dask Array

    Examples
    --------
    >>> from scipy import ndimage, misc
    >>> import numpy.fft
    >>> import matplotlib.pyplot as plt
    >>> fig, (ax1, ax2) = plt.subplots(1, 2)
    >>> plt.gray()  # show the filtered result in grayscale
    >>> ascent = misc.ascent()
    >>> image = numpy.fft.fft2(ascent)
    >>> result = ndimage.fourier_gaussian(image, sigma=4)
    >>> result = numpy.fft.ifft2(result)
    >>> ax1.imshow(ascent)
    """

    # Validate and normalize arguments
    image, sigma, n, axis = _utils._norm_args(image, sigma, n=n, axis=axis)

    # Compute frequencies
    ang_freq_grid = _utils._get_ang_freq_grid(
        image.shape,
        chunks=image.chunks,
        n=n,
        axis=axis,
        dtype=sigma.dtype
    )

    # Compute Fourier transformed Gaussian
    result = image.copy()
    scale = (sigma ** 2) / -2

    for ax, f in enumerate(ang_freq_grid):
        f *= f
        gaussian = da.exp(scale[ax] * f)
        gaussian = _utils._reshape_nd(gaussian, ndim=image.ndim, axis=ax)
        result *= gaussian

    return result


def fourier_shift(image, shift, n=-1, axis=-1):
    """
    Multi-dimensional fourier shift filter.

    The array is multiplied with the fourier transform of a shift operation.

    Parameters
    ----------
    image : array_like
        The input image.
    shift : float or sequence
        The size of the box used for filtering.
        If a float, `shift` is the same for all axes. If a sequence, `shift`
        has to contain one value for each axis.
    n : int, optional
        If `n` is negative (default), then the image is assumed to be the
        result of a complex fft.
        If `n` is larger than or equal to zero, the image is assumed to be the
        result of a real fft, and `n` gives the length of the array before
        transformation along the real transform direction.
    axis : int, optional
        The axis of the real transform.

    Returns
    -------
    fourier_shift : Dask Array

    Examples
    --------
    >>> from scipy import ndimage, misc
    >>> import matplotlib.pyplot as plt
    >>> import numpy.fft
    >>> fig, (ax1, ax2) = plt.subplots(1, 2)
    >>> plt.gray()  # show the filtered result in grayscale
    >>> ascent = misc.ascent()
    >>> image = numpy.fft.fft2(ascent)
    >>> result = ndimage.fourier_shift(image, shift=200)
    >>> result = numpy.fft.ifft2(result)
    >>> ax1.imshow(ascent)
    >>> ax2.imshow(result.real)  # the imaginary part is an artifact
    >>> plt.show()
    """

    if issubclass(image.dtype.type, numbers.Real):
        image = image.astype(complex)

    # Validate and normalize arguments
    image, shift, n, axis = _utils._norm_args(image, shift, n=n, axis=axis)

    # Constants with type converted
    J = image.dtype.type(1j)

    # Get the grid of frequencies
    ang_freq_grid = _utils._get_ang_freq_grid(
        image.shape,
        chunks=image.chunks,
        n=n,
        axis=axis,
        dtype=shift.dtype
    )

    # Apply shift
    result = image.copy()
    for ax, f in enumerate(ang_freq_grid):
        phase_shift = da.exp((-J) * shift[ax] * f)
        phase_shift = _utils._reshape_nd(phase_shift, ndim=image.ndim, axis=ax)
        result *= phase_shift

    return result


def fourier_uniform(image, size, n=-1, axis=-1):
    """
    Multi-dimensional uniform fourier filter.

    The array is multiplied with the fourier transform of a box of given
    size.

    Parameters
    ----------
    image : array_like
        The input image.
    size : float or sequence
        The size of the box used for filtering.
        If a float, `size` is the same for all axes. If a sequence, `size` has
        to contain one value for each axis.
    n : int, optional
        If `n` is negative (default), then the image is assumed to be the
        result of a complex fft.
        If `n` is larger than or equal to zero, the image is assumed to be the
        result of a real fft, and `n` gives the length of the array before
        transformation along the real transform direction.
    axis : int, optional
        The axis of the real transform.

    Returns
    -------
    fourier_uniform : Dask Array
        The filtered image. If `output` is given as a parameter, None is
        returned.

    Examples
    --------
    >>> from scipy import ndimage, misc
    >>> import numpy.fft
    >>> import matplotlib.pyplot as plt
    >>> fig, (ax1, ax2) = plt.subplots(1, 2)
    >>> plt.gray()  # show the filtered result in grayscale
    >>> ascent = misc.ascent()
    >>> image = numpy.fft.fft2(ascent)
    >>> result = ndimage.fourier_uniform(image, size=20)
    >>> result = numpy.fft.ifft2(result)
    >>> ax1.imshow(ascent)
    >>> ax2.imshow(result.real)  # the imaginary part is an artifact
    >>> plt.show()
    """

    # Validate and normalize arguments
    image, size, n, axis = _utils._norm_args(image, size, n=n, axis=axis)

    # Get the grid of frequencies
    freq_grid = _utils._get_freq_grid(
        image.shape,
        chunks=image.chunks,
        n=n,
        axis=axis,
        dtype=size.dtype
    )

    # Compute uniform filter
    result = image.copy()
    for ax, f in enumerate(freq_grid):
        uniform = da.sinc(size[ax] * f)
        uniform = _utils._reshape_nd(uniform, ndim=image.ndim, axis=ax)
        result *= uniform

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import dask.array as da
import numpy as np
import pytest

import dask_image.ndmorph

cupy = pytest.importorskip("cupy", minversion="9.0.0")


@pytest.fixture
def array():
    s = (10, 10)
    a = da.from_array(cupy.arange(int(np.prod(s)),
                      dtype=cupy.float32).reshape(s), chunks=5)
    return a


@pytest.mark.cupy
@pytest.mark.parametrize("func", [
    dask_image.ndmorph.binary_closing,
    dask_image.ndmorph.binary_dilation,
    dask_image.ndmorph.binary_erosion,
    dask_image.ndmorph.binary_opening,
])
def test_cupy_ndmorph(array, func):
    """Test convolve & correlate filters with cupy input arrays."""
    result = func(array)
    assert result.dtype == bool
    assert result._meta.dtype == bool
    assert isinstance(result._meta, cupy.ndarray)
    computed = result.compute()
    assert computed.dtype == bool
    assert isinstance(computed, cupy.ndarray)



# -*- coding: utf-8 -*-

import operator

import dask
import dask.array as da
import numpy as np
import scipy.ndimage
import scipy.sparse
import scipy.sparse.csgraph


def _get_ndimage_label_dtype():
    return scipy.ndimage.label([1, 0, 1])[0].dtype


LABEL_DTYPE = _get_ndimage_label_dtype()


def _get_connected_components_dtype():
    a = np.empty((0, 0), dtype=int)
    return scipy.sparse.csgraph.connected_components(a)[1].dtype


CONN_COMP_DTYPE = _get_connected_components_dtype()


def relabel_blocks(block_labeled, new_labeling):
    """
    Relabel a block-labeled array based on ``new_labeling``.

    Parameters
    ----------
    block_labeled : array of int
        The input label array.
    new_labeling : 1D array of int
        A new labeling, such that ``labeling[i] = j`` implies that
        any element in ``array`` valued ``i`` should be relabeled to ``j``.

    Returns
    -------
    relabeled : array of int, same shape as ``array``
        The relabeled input array.
    """
    new_labeling = new_labeling.astype(LABEL_DTYPE)
    relabeled = da.map_blocks(operator.getitem,
                              new_labeling,
                              block_labeled,
                              dtype=LABEL_DTYPE,
                              chunks=block_labeled.chunks)
    return relabeled


def _unique_axis(a, axis=0):
    """Find unique subarrays in axis in N-D array."""
    at = np.ascontiguousarray(a.swapaxes(0, axis))
    dt = np.dtype([("values", at.dtype, at.shape[1:])])
    atv = at.view(dt)
    r = np.unique(atv)["values"].swapaxes(0, axis)
    return r


def _across_block_label_grouping(face, structure):
    """
    Find a grouping of labels across block faces.

    We assume that the labels on either side of the block face are unique to
    that block. This is enforced elsewhere.

    Parameters
    ----------
    face : array-like
        This is the boundary, of thickness (2,), between two blocks.
    structure : array-like
        Structuring element for the labeling of the face. This should have
        length 3 along each axis and have the same number of dimensions as
        ``face``.

    Returns
    -------
    grouped : array of int, shape (2, M)
        If a column of ``grouped`` contains the values ``i`` and ``j``, it
        implies that labels ``i`` and ``j`` belong in the same group. These
        are edges in a global label connectivity graph.

    Examples
    --------
    >>> face = np.array([[1, 1, 0, 2, 2, 0, 8],
    ...                     [0, 7, 7, 7, 7, 0, 9]])
    >>> structure = np.ones((3, 3), dtype=bool)
    >>> _across_block_label_grouping(face, structure)
    array([[1, 2, 8],
           [2, 7, 9]], dtype=np.int32)

    This shows that 1-2 are connected, 2-7 are connected, and 8-9 are
    connected. The resulting graph is (1-2-7), (8-9).
    """
    common_labels = scipy.ndimage.label(face, structure)[0]
    matching = np.stack((common_labels.ravel(), face.ravel()), axis=1)
    unique_matching = _unique_axis(matching)
    valid = np.all(unique_matching, axis=1)
    unique_valid_matching = unique_matching[valid]
    common_labels, labels = unique_valid_matching.T
    in_group = np.flatnonzero(np.diff(common_labels) == 0)
    i = np.take(labels, in_group)
    j = np.take(labels, in_group + 1)
    grouped = np.stack((i, j), axis=0)
    return grouped


def _across_block_label_grouping_delayed(face, structure):
    """Delayed version of :func:`_across_block_label_grouping`."""
    _across_block_label_grouping_ = dask.delayed(_across_block_label_grouping)
    grouped = _across_block_label_grouping_(face, structure)
    return da.from_delayed(grouped, shape=(2, np.nan), dtype=LABEL_DTYPE)


@dask.delayed
def _to_csr_matrix(i, j, n):
    """Using i and j as coo-format coordinates, return csr matrix."""
    v = np.ones_like(i)
    mat = scipy.sparse.coo_matrix((v, (i, j)), shape=(n, n))
    return mat.tocsr()


def label_adjacency_graph(labels, structure, nlabels):
    """
    Adjacency graph of labels between chunks of ``labels``.

    Each chunk in ``labels`` has been labeled independently, and the labels
    in different chunks are guaranteed to be unique.

    Here we construct a graph connecting labels in different chunks that
    correspond to the same logical label in the global volume. This is true
    if the two labels "touch" across the block face as defined by the input
    ``structure``.

    Parameters
    ----------
    labels : dask array of int
        The input labeled array, where each chunk is independently labeled.
    structure : array of bool
        Structuring element, shape (3,) * labels.ndim.
    nlabels : delayed int
        The total number of labels in ``labels`` *before* correcting for
        global consistency.

    Returns
    -------
    mat : delayed scipy.sparse.csr_matrix
        This matrix has value 1 at (i, j) if label i is connected to
        label j in the global volume, 0 everywhere else.
    """

    if structure is None:
        structure = scipy.ndimage.generate_binary_structure(labels.ndim, 1)

    faces = _chunk_faces(labels.chunks, labels.shape, structure)
    all_mappings = [da.empty((2, 0), dtype=LABEL_DTYPE, chunks=1)]
    for face_slice in faces:
        face = labels[face_slice]
        mapped = _across_block_label_grouping_delayed(face, structure)
        all_mappings.append(mapped)
    all_mappings = da.concatenate(all_mappings, axis=1)
    i, j = all_mappings
    mat = _to_csr_matrix(i, j, nlabels + 1)
    return mat


def _chunk_faces(chunks, shape, structure):
    """
    Return slices for two-pixel-wide boundaries between chunks.

    Parameters
    ----------
    chunks : tuple of tuple of int
        The chunk specification of the array.
    shape : tuple of int
        The shape of the array.
    structure: array of bool
        Structuring element, shape (3,) * ndim.

    Returns
    -------
    faces : list of tuple of slices
        Each element in this list indexes a face between two chunks.

    Examples
    --------
    >>> import dask.array as da
    >>> import scipy.ndimage as ndi
    >>> a = da.arange(110, chunks=110).reshape((10, 11)).rechunk(5)
    >>> structure = ndi.generate_binary_structure(2, 1)
    >>> chunk_faces(a.chunks, a.shape, structure)
    [(slice(4, 6, None), slice(0, 5, None)),
     (slice(4, 6, None), slice(5, 10, None)),
     (slice(4, 6, None), slice(10, 11, None)),
     (slice(0, 5, None), slice(4, 6, None)),
     (slice(0, 5, None), slice(9, 11, None)),
     (slice(5, 10, None), slice(4, 6, None)),
     (slice(5, 10, None), slice(9, 11, None))]
    """

    ndim = len(shape)
    numblocks = tuple(list(len(c) for c in chunks))
    
    slices = da.core.slices_from_chunks(chunks)
    
    # arrange block/chunk indices on grid
    block_summary = np.arange(len(slices)).reshape(numblocks)
    
    faces = []
    for ind_curr_block, curr_block in enumerate(np.ndindex(numblocks)):
        
        for pos_structure_coord in np.array(np.where(structure)).T:
                        
            # only consider forward neighbors
            if min(pos_structure_coord) < 1 or \
               max(pos_structure_coord) < 2: continue

            neigh_block = [curr_block[dim] + pos_structure_coord[dim] - 1
                           for dim in range(ndim)]

            if max([neigh_block[dim] >= numblocks[dim] for dim in range(ndim)]): continue

            # get neighbor slice index
            ind_neigh_block = block_summary[tuple(neigh_block)]

            curr_slice = []
            for dim in range(ndim):
                # keep slice if not on boundary
                if slices[ind_curr_block][dim] == slices[ind_neigh_block][dim]:
                    curr_slice.append(slices[ind_curr_block][dim])
                # otherwise, add two-pixel-wide boundary
                else:
                    curr_slice.append(slice(
                        slices[ind_curr_block][dim].stop - 1,
                        slices[ind_curr_block][dim].stop + 1))
                    
            faces.append(tuple(curr_slice))
            
    return faces


def block_ndi_label_delayed(block, structure):
    """
    Delayed version of ``scipy.ndimage.label``.

    Parameters
    ----------
    block : dask array (single chunk)
        The input array to be labeled.
    structure : array of bool
        Structure defining the connectivity of the labeling.

    Returns
    -------
    labeled : dask array, same shape as ``block``.
        The labeled array.
    n : delayed int
        The number of labels in ``labeled``.
    """
    label = dask.delayed(scipy.ndimage.label, nout=2)
    labeled_block, n = label(block, structure=structure)
    n = dask.delayed(LABEL_DTYPE.type)(n)
    labeled = da.from_delayed(labeled_block, shape=block.shape,
                              dtype=LABEL_DTYPE)
    n = da.from_delayed(n, shape=(), dtype=LABEL_DTYPE)
    return labeled, n


def connected_components_delayed(csr_matrix):
    """
    Delayed version of scipy.sparse.csgraph.connected_components.

    This version only returns the (delayed) connected component labelling, not
    the number of components.
    """
    conn_comp = dask.delayed(scipy.sparse.csgraph.connected_components, nout=2)
    return da.from_delayed(conn_comp(csr_matrix, directed=False)[1],
                           shape=(np.nan,), dtype=CONN_COMP_DTYPE)



# -*- coding: utf-8 -*-

import functools
import operator
import numbers

import dask.array as da
import numpy as np


def _get_freq_grid(shape, chunks, axis, n, dtype=float):
    assert len(shape) == len(chunks)

    shape = tuple(shape)
    dtype = np.dtype(dtype).type

    assert (issubclass(dtype, numbers.Real) and
            not issubclass(dtype, numbers.Integral))

    axis = axis % len(shape)

    freq_grid = []
    for ax, (s, c) in enumerate(zip(shape, chunks)):
        if axis == ax and n > 0:
            f = da.fft.rfftfreq(n, chunks=c).astype(dtype)
        else:
            f = da.fft.fftfreq(s, chunks=c).astype(dtype)
        freq_grid.append(f)

    freq_grid = da.meshgrid(*freq_grid, indexing="ij", sparse=True)

    return freq_grid


def _get_ang_freq_grid(shape, chunks, axis, n, dtype=float):
    dtype = np.dtype(dtype).type

    assert (issubclass(dtype, numbers.Real) and
            not issubclass(dtype, numbers.Integral))

    pi = dtype(np.pi)

    freq_grid = _get_freq_grid(shape, chunks, axis, n, dtype=dtype)
    ang_freq_grid = tuple((2 * pi) * f for f in freq_grid)

    return ang_freq_grid


def _norm_args(a, s, n=-1, axis=-1):
    if issubclass(a.dtype.type, numbers.Integral):
        a = a.astype(float)

    if isinstance(s, numbers.Number):
        s = np.array(a.ndim * [s])
    elif not isinstance(s, da.Array):
        s = np.array(s)

    if issubclass(s.dtype.type, numbers.Integral):
        s = s.astype(a.real.dtype)
    elif not issubclass(s.dtype.type, numbers.Real):
        raise TypeError("The `s` must contain real value(s).")
    if s.shape != (a.ndim,):
        raise RuntimeError(
            "Shape of `s` must be 1-D and equal to the input's rank."
        )

    if n != -1 and a.shape[axis] != (n // 2 + 1):
        raise NotImplementedError(
            "In the case of real-valued images, it is required that "
            "(n // 2 + 1) == image.shape[axis]."
        )

    return (a, s, n, axis)


def _reshape_nd(arr, ndim, axis):
    """Promote a 1d array to ndim with non-singleton size along axis."""
    nd_shape = (1,) * axis + (arr.size,) + (1,) * (ndim - axis - 1)
    return arr.reshape(nd_shape)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest
import numpy as np
import scipy.ndimage

import dask.array as da

import dask_image.ndfilters


@pytest.mark.parametrize(
    "da_func",
    [
        dask_image.ndfilters.generic_filter,
    ],
)
@pytest.mark.parametrize(
    "err_type, function, size, footprint, origin",
    [
        (RuntimeError, lambda x: x, None, None, 0),
        (TypeError, lambda x: x, 1.0, None, 0),
        (RuntimeError, lambda x: x, (1,), None, 0),
        (RuntimeError, lambda x: x, [(1,)], None, 0),
        (RuntimeError, lambda x: x, 1, np.ones((1,)), 0),
        (RuntimeError, lambda x: x, None, np.ones((1,)), 0),
        (RuntimeError, lambda x: x, None, np.ones((1, 0)), 0),
        (RuntimeError, lambda x: x, 1, None, (0,)),
        (RuntimeError, lambda x: x, 1, None, [(0,)]),
        (ValueError, lambda x: x, 1, None, 1),
        (TypeError, lambda x: x, 1, None, 0.0),
        (TypeError, lambda x: x, 1, None, (0.0, 0.0)),
        (TypeError, lambda x: x, 1, None, 1 + 0j),
        (TypeError, lambda x: x, 1, None, (0 + 0j, 1 + 0j)),
    ],
)
def test_generic_filters_params(da_func, err_type, function, size, footprint,
                                origin):
    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    with pytest.raises(err_type):
        da_func(d, function, size=size, footprint=footprint, origin=origin)


@pytest.mark.parametrize(
    "da_func",
    [
        dask_image.ndfilters.generic_filter,
    ],
)
def test_generic_filter_shape_type(da_func):
    function = lambda x: x  # noqa: E731
    size = 1

    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    assert all([(type(s) is int) for s in d.shape])

    d2 = da_func(d, function, size=size)

    assert all([(type(s) is int) for s in d2.shape])


@pytest.mark.parametrize(
    "sp_func, da_func",
    [(scipy.ndimage.generic_filter, dask_image.ndfilters.generic_filter)],
)
@pytest.mark.parametrize(
    "function, size, footprint",
    [
        (lambda x: x, 1, None),
        (lambda x: x, (1, 1), None),
        (lambda x: x, None, np.ones((1, 1))),
    ],
)
def test_generic_filter_identity(sp_func, da_func, function, size, footprint):
    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    da.utils.assert_eq(d, da_func(d, function, size=size, footprint=footprint))

    da.utils.assert_eq(
        sp_func(a, function, size=size, footprint=footprint),
        da_func(d, function, size=size, footprint=footprint),
    )


@pytest.mark.parametrize(
    "da_func",
    [
        dask_image.ndfilters.generic_filter,
    ],
)
def test_generic_filter_comprehensions(da_func):
    da_wfunc = lambda arr: da_func(arr, lambda x: x, 1)  # noqa: E731

    np.random.seed(0)

    a = np.random.random((3, 12, 14))
    d = da.from_array(a, chunks=(3, 6, 7))

    l2s = [da_wfunc(d[i]) for i in range(len(d))]
    l2c = [da_wfunc(d[i])[None] for i in range(len(d))]

    da.utils.assert_eq(np.stack(l2s), da.stack(l2s))
    da.utils.assert_eq(np.concatenate(l2c), da.concatenate(l2c))


@pytest.mark.parametrize(
    "sp_func, da_func",
    [(scipy.ndimage.generic_filter, dask_image.ndfilters.generic_filter)],
)
@pytest.mark.parametrize(
    "function, size, footprint, origin",
    [
        (lambda x: (np.array(x) ** 2).sum(), 2, None, 0),
        (lambda x: (np.array(x) ** 2).sum(), None, np.ones((2, 3)), 0),
        (lambda x: (np.array(x) ** 2).sum(), None, np.ones((2, 3)), (0, 1)),
        (lambda x: (np.array(x) ** 2).sum(), None, np.ones((2, 3)), (0, -1)),
        (
            lambda x: (np.array(x) ** 2).sum(),
            None,
            (np.mgrid[-2: 2 + 1, -2: 2 + 1] ** 2).sum(axis=0) < 2.5 ** 2,
            0,
        ),
        (
            lambda x: (np.array(x) ** 2).sum(),
            None,
            (np.mgrid[-2: 2 + 1, -2: 2 + 1] ** 2).sum(axis=0) < 2.5 ** 2,
            (1, 2),
        ),
        (
            lambda x: (np.array(x) ** 2).sum(),
            None,
            (np.mgrid[-2: 2 + 1, -2: 2 + 1] ** 2).sum(axis=0) < 2.5 ** 2,
            (-1, -2),
        ),
        (lambda x: (np.array(x) ** 2).sum(), 5, None, 0),
        (lambda x: (np.array(x) ** 2).sum(), 7, None, 0),
        (lambda x: (np.array(x) ** 2).sum(), 8, None, 0),
        (lambda x: (np.array(x) ** 2).sum(), 10, None, 0),
        (lambda x: (np.array(x) ** 2).sum(), 5, None, 2),
        (lambda x: (np.array(x) ** 2).sum(), 5, None, -2),
    ],
)
def test_generic_filter_compare(sp_func, da_func, function, size, footprint,
                                origin):
    a = np.arange(140.0).reshape(10, 14)
    d = da.from_array(a, chunks=(5, 7))

    da.utils.assert_eq(
        sp_func(a, function, size=size, footprint=footprint, origin=origin),
        da_func(d, function, size=size, footprint=footprint, origin=origin),
    )



import dask.array as da
import numpy as np
from numpy.testing import assert_equal
import pytest

from dask_image.ndfilters import threshold_local


@pytest.mark.parametrize('block_size', [
    3,
    [3, 3],
    np.array([3, 3]),
    da.from_array(np.array([3, 3]), chunks=1),
    da.from_array(np.array([3, 3]), chunks=2),
])
class TestSimpleImage:
    def setup(self):
        self.image = da.from_array(np.array(
            [[0, 0, 1, 3, 5],
             [0, 1, 4, 3, 4],
             [1, 2, 5, 4, 1],
             [2, 4, 5, 2, 1],
             [4, 5, 1, 0, 0]], dtype=int), chunks=(5, 5))

    def test_threshold_local_gaussian(self, block_size):
        ref = np.array(
            [[False, False, False, False,  True],
             [False, False,  True, False,  True],
             [False, False,  True,  True, False],
             [False,  True,  True, False, False],
             [True,  True, False, False, False]]
        )
        out = threshold_local(self.image, block_size, method='gaussian')
        assert_equal(ref, (self.image > out).compute())

        out = threshold_local(self.image, block_size, method='gaussian',
                              param=1./3.)
        assert_equal(ref, (self.image > out).compute())

    def test_threshold_local_mean(self, block_size):
        ref = np.array(
            [[False, False, False, False,  True],
             [False, False,  True, False,  True],
             [False, False,  True,  True, False],
             [False,  True,  True, False, False],
             [True,  True, False, False, False]]
        )
        out = threshold_local(self.image, block_size, method='mean')
        assert_equal(ref, (self.image > out).compute())

    def test_threshold_local_median(self, block_size):
        ref = np.array(
            [[False, False, False, False,  True],
             [False, False,  True, False, False],
             [False, False,  True, False, False],
             [False, False,  True,  True, False],
             [False,  True, False, False, False]]
        )
        out = threshold_local(self.image, block_size, method='median')
        assert_equal(ref, (self.image > out).compute())


class TestGenericFilter:
    def setup(self):
        self.image = da.from_array(np.array(
            [[0, 0, 1, 3, 5],
             [0, 1, 4, 3, 4],
             [1, 2, 5, 4, 1],
             [2, 4, 5, 2, 1],
             [4, 5, 1, 0, 0]], dtype=int), chunks=(5, 5))

    def test_threshold_local_generic(self):
        ref = np.array(
            [[1.,  7., 16., 29., 37.],
             [5., 14., 23., 30., 30.],
             [13., 24., 30., 29., 21.],
             [25., 29., 28., 19., 10.],
             [34., 31., 23., 10.,  4.]]
        )
        unchanged = threshold_local(self.image, 1, method='generic', param=sum)
        out = threshold_local(self.image, 3, method='generic', param=sum)
        assert np.allclose(unchanged.compute(), self.image.compute())
        assert np.allclose(out.compute(), ref)

    def test_threshold_local_generic_invalid(self):
        expected_error_message = "Must include a valid function to use as "
        "the 'param' keyword argument."
        with pytest.raises(ValueError) as e:
            threshold_local(self.image, 3, method='generic', param='sum')
            assert e == expected_error_message


class TestInvalidArguments:
    def setup(self):
        self.image = da.from_array(np.array(
            [[0, 0, 1, 3, 5],
             [0, 1, 4, 3, 4],
             [1, 2, 5, 4, 1],
             [2, 4, 5, 2, 1],
             [4, 5, 1, 0, 0]], dtype=int), chunks=(5, 5))

    @pytest.mark.parametrize("method, block_size, error_type", [
        ('median', np.nan, TypeError),
    ])
    def test_nan_blocksize(self, method, block_size, error_type):
        with pytest.raises(error_type):
            threshold_local(self.image, block_size, method=method)

    def test_invalid_threshold_method(self):
        with pytest.raises(ValueError):
            threshold_local(self.image, 3, method='invalid')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from packaging import version

import dask
import dask.array as da
import numpy as np
import pytest
import scipy
import scipy.ndimage

import dask_image.ndinterp

# mode lists for the case with prefilter = False
_supported_modes = ['constant', 'nearest', 'reflect', 'mirror']
_unsupported_modes = ['wrap']

# additional modes are present in SciPy >= 1.6.0
if version.parse(scipy.__version__) >= version.parse('1.6.0'):
    _supported_modes += ['grid-constant', 'grid-mirror', 'grid-wrap']


def validate_spline_filter(n=2,
                           axis_size=64,
                           interp_order=3,
                           interp_mode='constant',
                           chunksize=32,
                           output=np.float64,
                           random_seed=0,
                           use_cupy=False,
                           axis=None,
                           input_as_non_dask_array=False,
                           depth=None):
    """
    Compare the outputs of `scipy.ndimage.spline_transform`
    and `dask_image.ndinterp.spline_transform`. If axis is not None, then
    `spline_transform1d` is tested instead.

    """
    if (
        np.dtype(output) != np.float64
        and version.parse(scipy.__version__) < version.parse('1.4.0')
    ):
        pytest.skip("bug in output dtype handling in SciPy < 1.4")

    # define test image
    np.random.seed(random_seed)
    image = np.random.random([axis_size] * n)

    if version.parse(dask.__version__) < version.parse("2020.1.0"):
        # older dask will fail if any chunks have size smaller than depth
        _depth = dask_image.ndinterp._get_default_depth(interp_order)
        rem = axis_size % chunksize
        if chunksize < _depth or (rem != 0 and rem < _depth):
            pytest.skip("older dask doesn't automatically rechunk")

    if input_as_non_dask_array:
        if use_cupy:
            import cupy as cp
            image_da = cp.asarray(image)
        else:
            image_da = image
    else:
        # transform into dask array
        image_da = da.from_array(image, chunks=[chunksize] * n)
        if use_cupy:
            import cupy as cp
            image_da = image_da.map_blocks(cp.asarray)

    if axis is not None:
        scipy_func = scipy.ndimage.spline_filter1d
        dask_image_func = dask_image.ndinterp.spline_filter1d
        kwargs = {'axis': axis}
    else:
        scipy_func = scipy.ndimage.spline_filter
        dask_image_func = dask_image.ndinterp.spline_filter
        kwargs = {}

    # transform with scipy
    image_t_scipy = scipy_func(
        image,
        output=output,
        order=interp_order,
        mode=interp_mode,
        **kwargs)

    # transform with dask-image
    image_t_dask = dask_image_func(
        image_da,
        output=output,
        order=interp_order,
        mode=interp_mode,
        depth=depth,
        **kwargs)
    image_t_dask_computed = image_t_dask.compute()

    rtol = atol = 1e-6
    out_dtype = np.dtype(output)
    assert image_t_scipy.dtype == image_t_dask_computed.dtype == out_dtype
    assert np.allclose(image_t_scipy, image_t_dask_computed,
                       rtol=rtol, atol=atol)


@pytest.mark.parametrize("n", [1, 2, 3])
@pytest.mark.parametrize("axis_size", [64])
@pytest.mark.parametrize("interp_order", range(2, 6))
@pytest.mark.parametrize("interp_mode", _supported_modes)
@pytest.mark.parametrize("chunksize", [32, 15])
def test_spline_filter_general(
    n,
    axis_size,
    interp_order,
    interp_mode,
    chunksize,
):

    validate_spline_filter(
        n=n,
        axis_size=axis_size,
        interp_order=interp_order,
        interp_mode=interp_mode,
        chunksize=chunksize,
        axis=None,
    )


@pytest.mark.cupy
@pytest.mark.parametrize("n", [2])
@pytest.mark.parametrize("axis_size", [32])
@pytest.mark.parametrize("interp_order", range(2, 6))
@pytest.mark.parametrize("interp_mode", _supported_modes[::2])
@pytest.mark.parametrize("chunksize", [16])
@pytest.mark.parametrize("axis", [None, -1])
@pytest.mark.parametrize("input_as_non_dask_array", [False, True])
def test_spline_filter_cupy(
    n,
    axis_size,
    interp_order,
    interp_mode,
    chunksize,
    axis,
    input_as_non_dask_array,
):

    pytest.importorskip("cupy", minversion="9.0.0")

    validate_spline_filter(
        n=n,
        axis_size=axis_size,
        interp_order=interp_order,
        interp_mode=interp_mode,
        chunksize=chunksize,
        axis=axis,
        input_as_non_dask_array=input_as_non_dask_array,
        use_cupy=True,
    )


@pytest.mark.parametrize("n", [1, 2, 3])
@pytest.mark.parametrize("axis_size", [48, 27])
@pytest.mark.parametrize("interp_order", range(2, 6))
@pytest.mark.parametrize("interp_mode", _supported_modes)
@pytest.mark.parametrize("chunksize", [33])
@pytest.mark.parametrize("axis", [0, 1, -1])
def test_spline_filter1d_general(
    n,
    axis_size,
    interp_order,
    interp_mode,
    chunksize,
    axis,
):
    if axis == 1 and n < 2:
        pytest.skip(msg="skip axis=1 for 1d signals")

    validate_spline_filter(
        n=n,
        axis_size=axis_size,
        interp_order=interp_order,
        interp_mode=interp_mode,
        chunksize=chunksize,
        axis=axis,
    )


@pytest.mark.parametrize("axis", [None, -1])
def test_spline_filter_non_dask_array_input(axis):

    validate_spline_filter(
        axis=axis,
        input_as_non_dask_array=True,
    )


@pytest.mark.parametrize("depth", [None, 24])
@pytest.mark.parametrize("axis", [None, -1])
def test_spline_filter_non_default_depth(depth, axis):

    validate_spline_filter(
        axis=axis,
        depth=depth,
    )


@pytest.mark.parametrize("depth", [(16, 32), [18]])
def test_spline_filter1d_invalid_depth(depth):

    with pytest.raises(ValueError):
        validate_spline_filter(
            axis=-1,
            depth=depth,
        )


@pytest.mark.parametrize("axis_size", [32])
@pytest.mark.parametrize("interp_order", range(2, 6))
@pytest.mark.parametrize("interp_mode", _unsupported_modes)
@pytest.mark.parametrize("axis", [None, -1])
def test_spline_filter_unsupported_modes(
    axis_size,
    interp_order,
    interp_mode,
    axis,
):

    with pytest.raises(NotImplementedError):
        validate_spline_filter(
            axis_size=axis_size,
            interp_order=interp_order,
            interp_mode=interp_mode,
            axis=axis,
        )


@pytest.mark.parametrize(
    "output", [np.float64, np.float32, "float32", np.dtype(np.float32)]
)
@pytest.mark.parametrize("axis", [None, -1])
def test_spline_filter_output_dtype(output, axis):

    validate_spline_filter(
        axis_size=32,
        interp_order=3,
        output=output,
        axis=axis,
    )


@pytest.mark.parametrize("axis", [None, -1])
def test_spline_filter_array_output_unsupported(axis):

    n = 2
    axis_size = 32
    shape = (axis_size,) * n

    with pytest.raises(TypeError):
        validate_spline_filter(
            n=n,
            axis_size=axis_size,
            interp_order=3,
            output=np.empty(shape),
            axis=axis,
        )
